"""Tests creating a custom extension"""

from collections.abc import Generator
from datetime import datetime
from typing import Any, Generic, TypeVar, cast

import pytest

import pystac
from pystac import (
    Asset,
    Catalog,
    Collection,
    Extent,
    Item,
    SpatialExtent,
    TemporalExtent,
)
from pystac.errors import ExtensionTypeError
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.serialization.identify import STACJSONDescription, STACVersionID

T = TypeVar("T", pystac.Catalog, pystac.Collection, pystac.Item, pystac.Asset)

SCHEMA_URI = "https://example.com/v2.0/custom-schema.json"

TEST_PROP = "test:prop"


class CustomExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Catalog | pystac.Collection | pystac.Item],
):
    def __init__(self, obj: pystac.STACObject | None) -> None:
        self.obj = obj

    def apply(self, test_prop: str | None) -> None:
        self.test_prop = test_prop

    @property
    def test_prop(self) -> str | None:
        return self._get_property(TEST_PROP, str)

    @test_prop.setter
    def test_prop(self, v: str | None) -> None:
        self._set_property(TEST_PROP, v)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> "CustomExtension[T]":
        if isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(CustomExtension[T], AssetCustomExtension(obj))
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(CustomExtension[T], ItemCustomExtension(obj))
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(CustomExtension[T], CollectionCustomExtension(obj))
        if isinstance(obj, pystac.Catalog):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(CustomExtension[T], CatalogCustomExtension(obj))

        raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class CatalogCustomExtension(CustomExtension[pystac.Catalog]):
    def __init__(self, catalog: pystac.Catalog) -> None:
        self.properties = catalog.extra_fields
        super().__init__(catalog)


class CollectionCustomExtension(CustomExtension[pystac.Collection]):
    def __init__(self, collection: pystac.Collection) -> None:
        self.properties = collection.extra_fields
        super().__init__(collection)


class ItemCustomExtension(CustomExtension[pystac.Item]):
    def __init__(self, item: pystac.Item) -> None:
        self.properties = item.properties
        super().__init__(item)


class AssetCustomExtension(CustomExtension[pystac.Asset]):
    def __init__(self, asset: pystac.Asset) -> None:
        self.properties = asset.extra_fields
        if asset.owner:
            if isinstance(asset.owner, pystac.Item):
                self.additional_read_properties = [asset.owner.properties]
            elif isinstance(asset.owner, pystac.Collection):
                self.additional_read_properties = [asset.owner.extra_fields]
        super().__init__(None)


class CustomExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: set[str] = {
        "custom",
        "https://example.com/v1.0/custom-schema.json",
    }
    stac_object_types: set[pystac.STACObjectType] = {
        pystac.STACObjectType.CATALOG,
        pystac.STACObjectType.COLLECTION,
        pystac.STACObjectType.ITEM,
    }

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        if version < "1.0.0-rc2" and info.object_type == pystac.STACObjectType.ITEM:
            if "test:old-prop-name" in obj["properties"]:
                obj["properties"][TEST_PROP] = obj["properties"]["test:old-prop-name"]
        super().migrate(obj, version, info)


@pytest.fixture
def add_extension_hooks() -> Generator[None]:
    pystac.EXTENSION_HOOKS.add_extension_hooks(CustomExtensionHooks())
    yield
    pystac.EXTENSION_HOOKS.remove_extension_hooks(SCHEMA_URI)


def test_add_to_item_asset(add_extension_hooks: None) -> None:
    item = Item("an-id", None, None, datetime.now(), {})
    item.add_asset("foo", Asset("http://pystac.test/asset.tif"))
    custom = CustomExtension.ext(item.assets["foo"], add_if_missing=True)
    assert CustomExtension.has_extension(item)
    custom.apply("bar")
    item_as_dict = item.to_dict()
    assert item_as_dict["assets"]["foo"]["test:prop"] == "bar"


def test_add_to_item(add_extension_hooks: None) -> None:
    item = Item("an-id", None, None, datetime.now(), {})
    custom = CustomExtension.ext(item, add_if_missing=True)
    assert CustomExtension.has_extension(item)
    custom.test_prop = "foo"
    item_as_dict = item.to_dict()
    assert item_as_dict["properties"]["test:prop"] == "foo"


def test_add_to_catalog(add_extension_hooks: None) -> None:
    catalog = Catalog("an-id", "a description")
    custom = CustomExtension.ext(catalog, add_if_missing=True)
    assert CustomExtension.has_extension(catalog)
    custom.test_prop = "foo"
    catalog_as_dict = catalog.to_dict()
    assert catalog_as_dict["test:prop"] == "foo"


def test_add_to_collection(add_extension_hooks: None) -> None:
    collection = Collection(
        "an-id",
        "a description",
        extent=Extent(
            spatial=SpatialExtent([-180.0, -90.0, 180.0, 90.0]),
            temporal=TemporalExtent([[datetime.now(), None]]),
        ),
    )
    custom = CustomExtension.ext(collection, add_if_missing=True)
    assert CustomExtension.has_extension(collection)
    custom.test_prop = "foo"
    collection_as_dict = collection.to_dict()
    assert collection_as_dict["test:prop"] == "foo"


def test_add_to_collection_asset(add_extension_hooks: None) -> None:
    collection = Collection(
        "an-id",
        "a description",
        extent=Extent(
            spatial=SpatialExtent([-180.0, -90.0, 180.0, 90.0]),
            temporal=TemporalExtent([[datetime.now(), None]]),
        ),
    )
    collection.add_asset("foo", Asset("http://pystac.test/asset.tif"))
    custom = CustomExtension.ext(collection.assets["foo"], add_if_missing=True)
    assert CustomExtension.has_extension(collection)
    custom.test_prop = "bar"
    collection_as_dict = collection.to_dict()
    assert collection_as_dict["assets"]["foo"]["test:prop"] == "bar"


def test_ext_non_stac_object(add_extension_hooks: None) -> None:
    with pytest.raises(ExtensionTypeError):
        CustomExtension.ext({})  # type: ignore


def test_migrates(add_extension_hooks: None) -> None:
    item = Item("an-id", None, None, datetime.now(), {})
    item_as_dict = item.to_dict()
    item_as_dict["stac_version"] = "1.0.0-rc.1"
    item_as_dict["stac_extensions"] = ["https://example.com/v1.0/custom-schema.json"]
    item_as_dict["properties"]["test:old-prop-name"] = "foo"
    item = Item.from_dict(item_as_dict, migrate=True)
    custom = CustomExtension.ext(item)
    assert custom.test_prop == "foo"



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import lru_cache
from typing import TYPE_CHECKING, Any

import pystac
from pystac.extensions.base import VERSION_REGEX
from pystac.serialization.identify import STACJSONDescription, STACVersionID

if TYPE_CHECKING:
    from pystac.stac_object import STACObject


class ExtensionHooks(ABC):
    @property
    @abstractmethod
    def schema_uri(self) -> str:
        """The schema_uri for the current version of this extension"""
        raise NotImplementedError

    @property
    @abstractmethod
    def prev_extension_ids(self) -> set[str]:
        """A set of previous extension IDs (schema URIs or old short ids)
        that should be migrated to the latest schema URI in the 'stac_extensions'
        property. Override with a class attribute so that the set of previous
        IDs is only created once.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def stac_object_types(self) -> set[pystac.STACObjectType]:
        """A set of STACObjectType for which migration logic will be applied."""
        raise NotImplementedError

    @lru_cache
    def _get_stac_object_types(self) -> set[str]:
        """Translation of stac_object_types to strings, cached"""
        return {x.value for x in self.stac_object_types}

    def get_object_links(self, obj: STACObject) -> list[str | pystac.RelType] | None:
        return None

    def has_extension(self, obj: dict[str, Any]) -> bool:
        schema_startswith = VERSION_REGEX.split(self.schema_uri)[0] + "/"
        return any(
            uri.startswith(schema_startswith) or uri in self.prev_extension_ids
            for uri in obj.get("stac_extensions", [])
        )

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        """Migrate a STAC Object in dict format from a previous version.
        The base implementation will update the stac_extensions to the latest
        schema ID. This method will only be called for STAC objects that have been
        identified as a previous version of STAC. Implementations should directly
        manipulate the obj dict. Remember to call super() in order to change out
        the old 'stac_extension' entry with the latest schema URI.
        """
        # Migrate schema versions
        for prev_id in self.prev_extension_ids:
            if prev_id in info.extensions:
                try:
                    i = obj["stac_extensions"].index(prev_id)
                    obj["stac_extensions"][i] = self.schema_uri
                except ValueError:
                    obj["stac_extensions"].append(self.schema_uri)
                break


class RegisteredExtensionHooks:
    hooks: dict[str, ExtensionHooks]

    def __init__(self, hooks: Iterable[ExtensionHooks]):
        self.hooks = {e.schema_uri: e for e in hooks}

    def add_extension_hooks(self, hooks: ExtensionHooks) -> None:
        e_id = hooks.schema_uri
        if e_id in self.hooks:
            raise pystac.ExtensionAlreadyExistsError(
                f"ExtensionDefinition with id '{e_id}' already exists."
            )

        self.hooks[e_id] = hooks

    def remove_extension_hooks(self, extension_id: str) -> None:
        if extension_id in self.hooks:
            del self.hooks[extension_id]

    def get_extended_object_links(self, obj: STACObject) -> list[str | pystac.RelType]:
        result: list[str | pystac.RelType] | None = None
        for ext in obj.stac_extensions:
            if ext in self.hooks:
                ext_result = self.hooks[ext].get_object_links(obj)
                if ext_result is not None:
                    if result is None:
                        result = ext_result
                    else:
                        result.extend(ext_result)
        return result or []

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        for hooks in self.hooks.values():
            if info.object_type in hooks._get_stac_object_types():
                hooks.migrate(obj, version, info)



"""Implements the :stac-ext:`Timestamps Extension <timestamps>`."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from typing import Any, Generic, Literal, TypeVar, cast

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.summaries import RangeSummary
from pystac.utils import datetime_to_str, map_opt, str_to_datetime

#: Generalized version of :class:`~pystac.Collection`, :class:`~pystac.Item`,
#: or :class:`~pystac.Asset`
T = TypeVar("T", pystac.Collection, pystac.Item, pystac.Asset)

SCHEMA_URI = "https://stac-extensions.github.io/timestamps/v1.1.0/schema.json"

PUBLISHED_PROP = "published"
EXPIRES_PROP = "expires"
UNPUBLISHED_PROP = "unpublished"


class TimestampsExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or :class:`~pystac.Asset` with properties from the
    :stac-ext:`Timestamps Extension <timestamps>`. This class is generic over the type
    of STAC Object to be extended (e.g. :class:`~pystac.Item`, :class:`~pystac.Asset`).

    To create a concrete instance of :class:`TimestampsExtension`, use the
    :meth:`TimestampsExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> ts_ext = TimestampsExtension.ext(item)
    """

    name: Literal["timestamps"] = "timestamps"

    def apply(
        self,
        published: datetime | None = None,
        expires: datetime | None = None,
        unpublished: datetime | None = None,
    ) -> None:
        """Applies timestamps extension properties to the extended Item.

        Args:
            published : Date and time the corresponding data
                was published the first time.
            expires : Date and time the corresponding data
                expires (is not valid any longer).
            unpublished : Date and time the corresponding data
                was unpublished.
        """
        self.published = published
        self.expires = expires
        self.unpublished = unpublished

    @property
    def published(self) -> datetime | None:
        """Gets or sets a datetime object that represents the date and time that the
        corresponding data was published the first time.

        'Published' has a different meaning depending on where it is used. If available
        in the asset properties, it refers to the timestamps valid for the actual data
        linked to the Asset Object. If it comes from the Item properties, it refers to
        timestamp valid for the metadata.
        """
        return map_opt(str_to_datetime, self._get_property(PUBLISHED_PROP, str))

    @published.setter
    def published(self, v: datetime | None) -> None:
        self._set_property(PUBLISHED_PROP, map_opt(datetime_to_str, v))

    @property
    def expires(self) -> datetime | None:
        """Gets or sets a datetime object that represents the date and time the
        corresponding data expires (is not valid any longer).

        'Expires' has a different meaning depending on where it is used. If
        available in the asset properties, it refers to the timestamps valid for the
        actual data linked to the Asset Object. If it comes from the Item properties,
        it refers to the timestamp valid for the metadata.
        """
        return map_opt(str_to_datetime, self._get_property(EXPIRES_PROP, str))

    @expires.setter
    def expires(self, v: datetime | None) -> None:
        self._set_property(EXPIRES_PROP, map_opt(datetime_to_str, v))

    @property
    def unpublished(self) -> datetime | None:
        """Gets or sets a datetime object that represents the date and time the
        corresponding data was unpublished.

        'Unpublished' has a different meaning depending on where it is used. If
        available in the asset properties, it refers to the timestamps valid for the
        actual data linked to the Asset Object. If it comes from the Item properties,
        it's referencing to the timestamp valid for the metadata.
        """
        return map_opt(str_to_datetime, self._get_property(UNPUBLISHED_PROP, str))

    @unpublished.setter
    def unpublished(self, v: datetime | None) -> None:
        self._set_property(UNPUBLISHED_PROP, map_opt(datetime_to_str, v))

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> TimestampsExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Timestamps
        Extension <timestamps>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(TimestampsExtension[T], ItemTimestampsExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(TimestampsExtension[T], AssetTimestampsExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesTimestampsExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesTimestampsExtension(obj)


class ItemTimestampsExtension(TimestampsExtension[pystac.Item]):
    """A concrete implementation of :class:`TimestampsExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Timestamps Extension <timestamps>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`TimestampsExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemTimestampsExtension Item id={self.item.id}>"


class AssetTimestampsExtension(TimestampsExtension[pystac.Asset]):
    """A concrete implementation of :class:`TimestampsExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties
    defined in the :stac-ext:`Timestamps Extension <timestamps>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`TimestampsExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetTimestampsExtension Asset href={self.asset_href}>"


class SummariesTimestampsExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Timestamps Extension <timestamps>`.
    """

    @property
    def published(self) -> RangeSummary[datetime] | None:
        """Get or sets the summary of :attr:`TimestampsExtension.published` values
        for this Collection.
        """

        return map_opt(
            lambda s: RangeSummary(
                str_to_datetime(s.minimum), str_to_datetime(s.maximum)
            ),
            self.summaries.get_range(PUBLISHED_PROP),
        )

    @published.setter
    def published(self, v: RangeSummary[datetime] | None) -> None:
        self._set_summary(
            PUBLISHED_PROP,
            map_opt(
                lambda s: RangeSummary(
                    datetime_to_str(s.minimum), datetime_to_str(s.maximum)
                ),
                v,
            ),
        )

    @property
    def expires(self) -> RangeSummary[datetime] | None:
        """Get or sets the summary of :attr:`TimestampsExtension.expires` values
        for this Collection.
        """

        return map_opt(
            lambda s: RangeSummary(
                str_to_datetime(s.minimum), str_to_datetime(s.maximum)
            ),
            self.summaries.get_range(EXPIRES_PROP),
        )

    @expires.setter
    def expires(self, v: RangeSummary[datetime] | None) -> None:
        self._set_summary(
            EXPIRES_PROP,
            map_opt(
                lambda s: RangeSummary(
                    datetime_to_str(s.minimum), datetime_to_str(s.maximum)
                ),
                v,
            ),
        )

    @property
    def unpublished(self) -> RangeSummary[datetime] | None:
        """Get or sets the summary of :attr:`TimestampsExtension.unpublished` values
        for this Collection.
        """

        return map_opt(
            lambda s: RangeSummary(
                str_to_datetime(s.minimum), str_to_datetime(s.maximum)
            ),
            self.summaries.get_range(UNPUBLISHED_PROP),
        )

    @unpublished.setter
    def unpublished(self, v: RangeSummary[datetime] | None) -> None:
        self._set_summary(
            UNPUBLISHED_PROP,
            map_opt(
                lambda s: RangeSummary(
                    datetime_to_str(s.minimum), datetime_to_str(s.maximum)
                ),
                v,
            ),
        )


class TimestampsExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "timestamps",
        "https://stac-extensions.github.io/timestamps/v1.0.0/schema.json",
    }
    stac_object_types = {pystac.STACObjectType.ITEM}


TIMESTAMPS_EXTENSION_HOOKS: ExtensionHooks = TimestampsExtensionHooks()



from __future__ import annotations

from copy import copy, deepcopy
from typing import TYPE_CHECKING, Any, TypeVar, cast

import pystac
from pystac import RelType, STACError, STACObjectType
from pystac.asset import Asset, Assets
from pystac.catalog import Catalog
from pystac.collection import Collection
from pystac.errors import DeprecatedWarning, ExtensionNotImplemented
from pystac.link import Link
from pystac.serialization import (
    identify_stac_object,
    identify_stac_object_type,
    migrate_to_latest,
)
from pystac.stac_object import STACObject
from pystac.utils import (
    datetime_to_str,
    is_absolute_href,
    make_absolute_href,
    make_relative_href,
    str_to_datetime,
)

#: Generalized version of :class:`Item`
T = TypeVar("T", bound="Item")

if TYPE_CHECKING:
    # avoids conflicts since there are also kwargs and attrs called `datetime`
    from datetime import datetime as Datetime

    from pystac.extensions.ext import ItemExt


class Item(STACObject, Assets):
    """An Item is the core granular entity in a STAC, containing the core metadata
    that enables any client to search or crawl online catalogs of spatial 'assets' -
    satellite imagery, derived data, DEM's, etc.

    Args:
        id : Provider identifier. Must be unique within the STAC.
        geometry : Defines the full footprint of the asset represented by this
            item, formatted according to
            `RFC 7946, section 3.1 (GeoJSON) <https://tools.ietf.org/html/rfc7946>`_.
        bbox :  Bounding Box of the asset represented by this item
            using either 2D or 3D geometries. The length of the array must be 2*n
            where n is the number of dimensions. Could also be None in the case of a
            null geometry.
        datetime : datetime associated with this item. If None,
            a start_datetime and end_datetime must be supplied.
        properties : A dictionary of additional metadata for the item.
        start_datetime : Optional inclusive start datetime, part of common metadata.
            This value will override any `start_datetime` key in properties.
        end_datetime : Optional inclusive end datetime, part of common metadata.
            This value will override any `end_datetime` key in properties.
        stac_extensions : Optional list of extensions the Item implements.
        href : Optional HREF for this item, which be set as the item's
            self link's HREF.
        collection : The Collection or Collection ID that this item
            belongs to.
        extra_fields : Extra fields that are part of the top-level JSON
            properties of the Item.
        assets : A dictionary mapping string keys to :class:`~pystac.Asset` objects. All
            :class:`~pystac.Asset` values in the dictionary will have their
            :attr:`~pystac.Asset.owner` attribute set to the created Item.
    """

    assets: dict[str, Asset]
    """Dictionary of :class:`~pystac.Asset` objects, each with a unique key."""

    bbox: list[float] | None
    """Bounding Box of the asset represented by this item using either 2D or 3D
    geometries. The length of the array is 2*n where n is the number of dimensions.
    Could also be None in the case of a null geometry."""

    collection: Collection | None
    """:class:`~pystac.Collection` to which this Item belongs, if any."""

    collection_id: str | None
    """The Collection ID that this item belongs to, if any."""

    datetime: Datetime | None
    """Datetime associated with this item. If ``None``, then
    :attr:`~pystac.CommonMetadata.start_datetime` and
    :attr:`~pystac.CommonMetadata.end_datetime` in :attr:`~pystac.Item.common_metadata`
    will supply the datetime range of the Item."""

    extra_fields: dict[str, Any]
    """Extra fields that are part of the top-level JSON fields the Item."""

    geometry: dict[str, Any] | None
    """Defines the full footprint of the asset represented by this item, formatted
    according to `RFC 7946, section 3.1 (GeoJSON)
    <https://tools.ietf.org/html/rfc7946>`_."""

    id: str
    """Provider identifier. Unique within the STAC."""

    links: list[Link]
    """A list of :class:`~pystac.Link` objects representing all links associated with
    this Item."""

    properties: dict[str, Any]
    """A dictionary of additional metadata for the Item."""

    stac_extensions: list[str]
    """List of extensions the Item implements."""

    STAC_OBJECT_TYPE = STACObjectType.ITEM

    def __init__(
        self,
        id: str,
        geometry: dict[str, Any] | None,
        bbox: list[float] | None,
        datetime: Datetime | None,
        properties: dict[str, Any],
        start_datetime: Datetime | None = None,
        end_datetime: Datetime | None = None,
        stac_extensions: list[str] | None = None,
        href: str | None = None,
        collection: str | Collection | None = None,
        extra_fields: dict[str, Any] | None = None,
        assets: dict[str, Asset] | None = None,
    ):
        super().__init__(stac_extensions or [])

        self.id = id
        self.geometry = geometry
        self.bbox = bbox
        self.properties = properties
        if extra_fields is None:
            self.extra_fields = {}
        else:
            self.extra_fields = extra_fields

        self.assets: dict[str, Asset] = {}

        self.datetime: Datetime | None = None
        if start_datetime:
            properties["start_datetime"] = datetime_to_str(start_datetime)
        if end_datetime:
            properties["end_datetime"] = datetime_to_str(end_datetime)
        if datetime is None:
            if "start_datetime" not in properties or "end_datetime" not in properties:
                raise STACError(
                    "Invalid Item: If datetime is None, "
                    "a start_datetime and end_datetime "
                    "must be supplied."
                )
            self.datetime = None
        else:
            self.datetime = datetime

        if href is not None:
            self.set_self_href(href)

        self.collection_id: str | None = None
        if collection is None:
            self.collection = None
        else:
            if isinstance(collection, Collection):
                self.set_collection(collection)
            else:
                self.collection_id = collection

        self.assets = {}
        if assets is not None:
            for k, asset in assets.items():
                self.add_asset(k, asset)

    def __repr__(self) -> str:
        return f"<Item id={self.id}>"

    def __getstate__(self) -> dict[str, Any]:
        """Ensure that pystac does not encode too much information when pickling"""
        d = self.__dict__.copy()

        d["links"] = [
            (
                link.to_dict(transform_href=False)
                if link.get_href(transform_href=False)
                else link
            )
            for link in d["links"]
        ]

        return d

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Ensure that pystac knows how to decode the pickled object"""
        d = state.copy()

        d["links"] = [
            Link.from_dict(link).set_owner(self) if isinstance(link, dict) else link
            for link in d["links"]
        ]

        self.__dict__ = d

    def set_self_href(self, href: str | None) -> None:
        """Sets the absolute HREF that is represented by the ``rel == 'self'``
        :class:`~pystac.Link`.

        Changing the self HREF of the item will ensure that all asset HREFs
        remain valid. If asset HREFs are relative, the HREFs will change
        to point to the same location based on the new item self HREF,
        either by making them relative to the new location or making them
        absolute links if the new location does not share the same protocol
        as the old location.

        Args:
            href : The absolute HREF of this object. If the given HREF
                is not absolute, it will be transformed to an absolute
                HREF based on the current working directory. If this is None
                the call will clear the self HREF link.
        """
        prev_href = self.get_self_href()
        super().set_self_href(href)
        new_href = self.get_self_href()  # May have been made absolute.

        if prev_href is not None and new_href is not None:
            # Make sure relative asset links remain valid.
            for asset in self.assets.values():
                asset_href = asset.href
                if not is_absolute_href(asset_href):
                    abs_href = make_absolute_href(asset_href, prev_href)
                    new_relative_href = make_relative_href(abs_href, new_href)
                    asset.href = new_relative_href

    def get_datetime(self, asset: Asset | None = None) -> Datetime | None:
        """Gets an Item or an Asset datetime.

        If an Asset is supplied and the Item property exists on the Asset,
        returns the Asset's value. Otherwise returns the Item's value.

        Returns:
            datetime or None
        """
        if asset is None or "datetime" not in asset.extra_fields:
            return self.datetime
        else:
            asset_dt = asset.extra_fields.get("datetime")
            if asset_dt is None:
                return None
            else:
                return str_to_datetime(asset_dt)

    def set_datetime(self, datetime: Datetime, asset: Asset | None = None) -> None:
        """Set an Item or an Asset datetime.

        If an Asset is supplied, sets the property on the Asset.
        Otherwise sets the Item's value.
        """
        if asset is None:
            self.datetime = datetime
        else:
            asset.extra_fields["datetime"] = datetime_to_str(datetime)

    def set_collection(self, collection: Collection | None) -> Item:
        """Set the collection of this item.

        This method will replace any existing Collection link and attribute for
        this item.

        Args:
            collection : The collection to set as this
                item's collection. If None, will clear the collection.

        Returns:
            Item: self
        """
        self.remove_links(pystac.RelType.COLLECTION)
        self.collection_id = None
        if collection is not None:
            self.add_link(Link.collection(collection))
            self.collection_id = collection.id

        return self

    def get_collection(self) -> Collection | None:
        """Gets the collection of this item, if one exists.

        Returns:
            Collection or None: If this item belongs to a collection, returns
            a reference to the collection. Otherwise returns None.
        """
        collection_link = self.get_single_link(pystac.RelType.COLLECTION)
        if collection_link is None:
            return None
        else:
            return cast(
                Collection, collection_link.resolve_stac_object(self.get_root()).target
            )

    def add_derived_from(self, *items: Item | str) -> Item:
        """Add one or more items that this is derived from.

        This method will add to any existing "derived_from" links.

        Args:
            items : Items (or href of items) to add to derived_from links.

        Returns:
            Item: self
        """
        for item in items:
            self.add_link(Link.derived_from(item))
        return self

    def remove_derived_from(self, item_id: str) -> None:
        """Remove an item that this is derived from.

        This method will remove from existing "derived_from" links.

        Args:
            item_id : ID of item to remove from derived_from links.
        """
        new_links: list[pystac.Link] = []

        for link in self.links:
            if link.rel != pystac.RelType.DERIVED_FROM:
                new_links.append(link)
            else:
                try:
                    item = cast(Item, link.resolve_stac_object().target)
                except Exception as e:
                    raise pystac.STACError(
                        "Link failed to resolve. Use remove_links instead."
                    ) from e
                if item.id != item_id:
                    new_links.append(link)
        self.links = new_links

    def get_derived_from(self) -> list[Item]:
        """Get the items that this is derived from.

        Returns:
            List[Item]: Returns a reference to the derived_from items.
        """
        links = self.get_links(pystac.RelType.DERIVED_FROM)
        try:
            return [cast(Item, link.resolve_stac_object().target) for link in links]
        except Exception as e:
            raise pystac.STACError(
                "Link failed to resolve. Use get_links instead."
            ) from e

    def to_dict(
        self, include_self_link: bool = True, transform_hrefs: bool = True
    ) -> dict[str, Any]:
        links = self.links
        if not include_self_link:
            links = [x for x in links if x.rel != pystac.RelType.SELF]

        assets = {k: v.to_dict() for k, v in self.assets.items()}

        if self.datetime is not None:
            self.properties["datetime"] = datetime_to_str(self.datetime)
        else:
            self.properties["datetime"] = None

        d: dict[str, Any] = {
            "type": "Feature",
            "stac_version": pystac.get_stac_version(),
            "stac_extensions": self.stac_extensions if self.stac_extensions else [],
            "id": self.id,
            "geometry": self.geometry,
            "bbox": self.bbox if self.bbox is not None else [],
            "properties": self.properties,
            "links": [link.to_dict(transform_href=transform_hrefs) for link in links],
            "assets": assets,
        }

        if self.collection_id:
            d["collection"] = self.collection_id

        for key in self.extra_fields:
            d[key] = self.extra_fields[key]

        # This field is prohibited if there's no geometry
        if not self.geometry:
            d.pop("bbox")

        return d

    def clone(self) -> Item:
        cls = self.__class__
        clone = cls(
            id=self.id,
            geometry=deepcopy(self.geometry),
            bbox=copy(self.bbox),
            datetime=copy(self.datetime),
            properties=deepcopy(self.properties),
            stac_extensions=deepcopy(self.stac_extensions),
            collection=self.collection_id,
            assets={k: asset.clone() for k, asset in self.assets.items()},
            extra_fields=deepcopy(self.extra_fields),
        )
        for link in self.links:
            clone.add_link(link.clone())

        return clone

    def _object_links(self) -> list[str | pystac.RelType]:
        return [
            pystac.RelType.COLLECTION,
            *pystac.EXTENSION_HOOKS.get_extended_object_links(self),
        ]

    @classmethod
    def from_dict(
        cls: type[T],
        d: dict[str, Any],
        href: str | None = None,
        root: Catalog | None = None,
        migrate: bool = True,
        preserve_dict: bool = True,
    ) -> T:
        import warnings

        from pystac.extensions.version import ItemVersionExtension

        if preserve_dict:
            d = deepcopy(d)

        if migrate:
            info = identify_stac_object(d)
            d = migrate_to_latest(d, info)

        if not cls.matches_object_type(d):
            raise pystac.STACTypeError(d, cls)

        # some fields are passed through to __init__
        pass_through_fields = [
            "id",
            "geometry",
            "bbox",
            "stac_extensions",
            "collection",
        ]

        # some fields need some parsing or special handling
        parse_fields = ["links", "assets", "properties"]
        links = d.get("links", [])
        assets = d.get("assets", {})
        properties = d.get("properties", {})
        datetime = properties.get("datetime")
        if datetime is not None:
            datetime = str_to_datetime(datetime)

        # some fields are excluded from __init__ entirely
        exclude_fields = ["type", "stac_version"]

        # get all the fields that are not passed through, parsed, or excluded
        extra_fields = {
            k: v
            for k, v in d.items()
            if k not in [*pass_through_fields, *parse_fields, *exclude_fields]
        }

        item = cls(
            **{k: d.get(k) for k in pass_through_fields},  # type: ignore
            datetime=datetime,
            properties=properties,
            extra_fields=extra_fields,
            href=href,
            assets={k: Asset.from_dict(v) for k, v in assets.items()},
        )

        for link in links:
            if href is None or link.get("rel", None) != RelType.SELF:
                item.add_link(Link.from_dict(link))

        if root:
            item.set_root(root)

        try:
            version = ItemVersionExtension.ext(item)
            if version.deprecated:
                warnings.warn(
                    f"The item '{item.id}' is deprecated.",
                    DeprecatedWarning,
                )
            # Item asset deprecation checks pending version extension support
        except ExtensionNotImplemented:
            pass

        return item

    @property
    def common_metadata(self) -> pystac.CommonMetadata:
        """Access the item's common metadata fields as a
        :class:`~pystac.CommonMetadata` object."""
        return pystac.CommonMetadata(self)

    def full_copy(
        self, root: Catalog | None = None, parent: Catalog | None = None
    ) -> Item:
        return cast(Item, super().full_copy(root, parent))

    @classmethod
    def matches_object_type(cls, d: dict[str, Any]) -> bool:
        for field in ("type", "stac_version"):
            if field not in d:
                raise pystac.STACTypeError(d, cls, f"'{field}' is missing.")
        return identify_stac_object_type(d) == STACObjectType.ITEM

    @property
    def __geo_interface__(self) -> dict[str, Any]:
        """Returns this item as a dictionary.

        This just calls `to_dict` without self links or transforming any hrefs.

        https://gist.github.com/sgillies/2217756

        Returns:
            Dict[str, Any]: This item as a dictionary.
        """
        return self.to_dict(include_self_link=False, transform_hrefs=False)

    @property
    def ext(self) -> ItemExt:
        """Accessor for extension classes on this item

        Example::

            item.ext.proj.code = "EPSG:4326"
        """
        from pystac.extensions.ext import ItemExt

        return ItemExt(stac_object=self)



"""Implements the :stac-ext:`Machine Learning Model (MLM) Extension <mlm>`.

This documentation does not provide a full-detail description of the meaning of each
parameter and how to use them. For an in-depth description of this extension, and the
use of each property, please refer to the official
:stac-ext:`documentation <mlm>`
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Iterable
from typing import Any, Generic, Literal, TypeVar, cast

import pystac
from pystac.errors import STACError
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
)
from pystac.extensions.classification import Classification
from pystac.extensions.hooks import ExtensionHooks
from pystac.extensions.raster import DataType
from pystac.serialization.identify import STACJSONDescription, STACVersionID
from pystac.utils import StringEnum, get_required

#: Generalized version of :class:`pystac.Item`, :class:`pystac.ItemAssetDefinition`,
#: :class:`pystac.Collection`, or :class:`pystac.Asset`
T = TypeVar(
    "T", pystac.Item, pystac.ItemAssetDefinition, pystac.Collection, pystac.Asset
)
AssetExtensionType = TypeVar("AssetExtensionType", bound="_AssetMLMExtension")

SCHEMA_URI_PATTERN: str = "https://stac-extensions.github.io/mlm/v{version}/schema.json"
DEFAULT_VERSION: str = "1.4.0"
SUPPORTED_VERSIONS: list[str] = ["1.0.0", "1.1.0", "1.2.0", "1.3.0", "1.4.0"]

PREFIX: str = "mlm:"

# Model Band Object properties
NAME_MODEL_BAND_OBJECT_PROP = "name"
FORMAT_MODEL_BAND_OBJECT_PROP = "format"
EXPRESSION_MODEL_BAND_OBJECT_PROP = "expression"

# Input Structure properties
SHAPE_INPUT_STRUCTURE_PROP = "shape"
DIM_ORDER_INPUT_STRUCTURE_PROP = "dim_order"
DATA_TYPE_INPUT_STRUCTURE_PROP = "data_type"

# Field names: Model Input Object
NAME_INPUT_OBJECT_PROP: str = "name"
BANDS_INPUT_OBJECT_PROP: str = "bands"
INPUT_INPUT_OBJECT_PROP: str = "input"
DESCRIPTION_INPUT_OBJECT_PROP: str = "description"
VALUE_SCALING_INPUT_OBJECT_PROP: str = "value_scaling"
RESIZE_TYPE_INPUT_OBJECT_PROP: str = "resize_type"
PRE_PROCESSING_FUNCTION_INPUT_OBJECT_PROP: str = "pre_processing_function"

# Output Structure properties
SHAPE_RESULT_STRUCTURE_PROP = "shape"
DIM_ORDER_RESULT_STRUCTURE_PROP = "dim_order"
DATA_TYPE_RESULT_STRUCTURE_PROP = "data_type"

# ProcessingExpression fields
FORMAT_PROCESSING_EXPRESSION_PROP = "format"
EXPRESSION_PROCESSING_EXPRESSION_PROP = "expression"

# ValueScaling fields
TYPE_VALUE_SCALING_PROP = "type"
MINIMUM_VALUE_SCALING_PROP = "minimum"
MAXIMUM_VALUE_SCALING_PROP = "maximum"
MEAN_VALUE_SCALING_PROP = "mean"
STDDEV_VALUE_SCALING_PROP = "stddev"
VALUE_VALUE_SCALING_PROP = "value"
FORMAT_VALUE_SCALING_PROP = "format"
EXPRESSION_VALUE_SCALING_PROP = "expression"

# Output properties
NAME_RESULT_PROP = "name"
TASKS_RESULT_PROP = "tasks"
RESULT_RESULT_PROP = "result"
DESCRIPTION_RESULT_PROP = "description"
CLASSES_RESULT_PROP = "classification:classes"
POST_PROCESSING_FUNCTION_RESULT_PROP = "post_processing_function"

# Field names
NAME_PROP: str = PREFIX + "name"
ARCHITECTURE_PROP: str = PREFIX + "architecture"
TASKS_PROP: str = PREFIX + "tasks"
FRAMEWORK_PROP: str = PREFIX + "framework"
FRAMEWORK_VERSION_PROP: str = PREFIX + "framework_version"
MEMORY_SIZE_PROP: str = PREFIX + "memory_size"
TOTAL_PARAMETERS_PROP: str = PREFIX + "total_parameters"
PRETRAINED_PROP: str = PREFIX + "pretrained"
PRETRAINED_SOURCE_PROP: str = PREFIX + "pretrained_source"
BATCH_SIZE_SUGGESTION_PROP: str = PREFIX + "batch_size_suggestion"
ACCELERATOR_PROP: str = PREFIX + "accelerator"
ACCELERATOR_CONSTRAINED_PROP: str = PREFIX + "accelerator_constrained"
ACCELERATOR_SUMMARY_PROP: str = PREFIX + "accelerator_summary"
ACCELERATOR_COUNT_PROP: str = PREFIX + "accelerator_count"
INPUT_PROP: str = PREFIX + "input"
OUTPUT_PROP: str = PREFIX + "output"
HYPERPARAMETERS_PROP: str = PREFIX + "hyperparameters"

ARTIFACT_TYPE_ASSET_PROP = PREFIX + "artifact_type"
COMPILE_METHOD_ASSET_PROP = PREFIX + "compile_method"
ENTRYPOINT_ASSET_PROP = PREFIX + "entrypoint"


class TaskType(StringEnum):
    """
    An enumeration of Tasks supported by the extension
    """

    REGRESSION = "regression"
    CLASSIFICATION = "classification"
    SCENE_CLASSIFICATION = "scene-classification"
    DETECTION = "detection"
    OBJECT_DETECTION = "object-detection"
    SEGMENTATION = "segmentation"
    SEMANTIC_SEGMENTATION = "semantic-segmentation"
    INSTANCE_SEGMENTATION = "instance-segmentation"
    PANOPTIC_SEGMENTATION = "panoptic-segmentation"
    SIMILARITY_SEARCH = "similarity-search"
    GENERATIVE = "generative"
    IAMGE_CAPTIONING = "image-captioning"
    SUPER_RESOLUTION = "super-resolution"


class AcceleratorType(StringEnum):
    """
    An enumeration of accelerators supported by the extension
    """

    AMD64 = "amd64"
    CUDA = "cuda"
    XLA = "xla"
    AMD_ROCM = "amd-rocm"
    INTEL_IPEX_CPU = "intel-ipex-cpu"
    INTEL_IPEX_GPU = "intel-ipex-gpu"
    MACOS_ARM = "macos-arm"


class ResizeType(StringEnum):
    """
    An enumeration of Resize operations supported by the extension
    """

    CROP = "crop"
    PAD = "pad"
    INTERPOLATION_NEAREST = "interpolate-nearest"
    INTERPOLATION_LINEAR = "interpolate-linear"
    INTERPOLATION_CUBIC = "interpolation-cubic"
    INTERPOLATION_AREA = "interpolation-area"
    INTERPOLATION_LANCZOS4 = "interpolation-lanczos4"
    INTERPOLATION_MAX = "interpolation-max"
    WRAP_FILL_OUTLIERS = "wrap-fill-outliers"
    WRAP_INVERSE_MAP = "wrap-inverse-map"


class ValueScalingType(StringEnum):
    """
    An enumeratino of Value Scaling operations supported by the extension
    """

    MIN_MAX = "min-max"
    Z_SCORE = "z-score"
    CLIP = "clip"
    CLIP_MIN = "clip-min"
    CLIP_MAX = "clip-max"
    OFFSET = "offset"
    SCALE = "scale"
    PROCESSING = "processing"


class ModelBand:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelBand):
            raise NotImplementedError
        return (
            self.name == other.name
            and self.format == other.format
            and self.expression == other.expression
        )

    def apply(
        self, name: str, format: str | None = None, expression: Any | None = None
    ) -> None:
        """
        Set the properties for a new ModelBand.

        Args:
            name: Name of the band referring to an extended band definition
            format: The type of expression that is specified in the expression property
            expression: An expression compliant with the format specified.
                The cxpression can be applied to any data type and depends on the format
                given
        """
        self.name = name
        self.format = format
        self.expression = expression

    @classmethod
    def create(
        cls, name: str, format: str | None = None, expression: Any | None = None
    ) -> ModelBand:
        """
        Create a new ModelBand.

        Args:
            name: Name of the band referring to an extended band definition
            format: The type of expression that is specified in the expression property
            expression: An expression compliant with the format specified.
                The expression can be applied to any data type and depends on the
                format given
        """
        c = cls({})
        c.apply(name=name, format=format, expression=expression)
        return c

    @property
    def name(self) -> str:
        """
        Get or set the required name property of a ModelBand object
        """
        return cast(
            str,
            get_required(
                self.properties.get(NAME_MODEL_BAND_OBJECT_PROP),
                self,
                NAME_MODEL_BAND_OBJECT_PROP,
            ),
        )

    @name.setter
    def name(self, v: str) -> None:
        self.properties[NAME_MODEL_BAND_OBJECT_PROP] = v

    @property
    def format(self) -> str | None:
        """
        Get or set the optional format property of a ModelBand object
        """
        return self.properties.get(FORMAT_MODEL_BAND_OBJECT_PROP)

    @format.setter
    def format(self, v: str | None) -> None:
        if v is not None:
            self.properties[FORMAT_MODEL_BAND_OBJECT_PROP] = v
        else:
            self.properties.pop(FORMAT_MODEL_BAND_OBJECT_PROP, None)

    @property
    def expression(self) -> Any:
        """
        Get or set the optional expression property of a ModelBand object
        """
        return self.properties.get(EXPRESSION_MODEL_BAND_OBJECT_PROP)

    @expression.setter
    def expression(self, v: Any) -> None:
        if v is not None:
            self.properties[EXPRESSION_MODEL_BAND_OBJECT_PROP] = v
        else:
            self.properties.pop(EXPRESSION_MODEL_BAND_OBJECT_PROP, None)

    def to_dict(self) -> dict[str, Any]:
        """
        Returns the dictionary encoding of this ModelBand object

        Returns:
            dict[str, Any
        """
        return self.properties


class ProcessingExpression:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ProcessingExpression):
            raise NotImplementedError
        else:
            return self.format == other.format and self.expression == other.expression

    def apply(self, format: str, expression: Any) -> None:
        """
        Set the properties for a new ProcessingExpression

        Args:
            format: The type of the expression that is specified in the expression
                property.
            expression: An expression compliant with the format specified. The
                expression can be any data type and depends on the format given,
                e.g. string or object.
        """
        self.format = format
        self.expression = expression

    @classmethod
    def create(cls, format: str, expression: Any) -> ProcessingExpression:
        """
        Creates a new ProcessingExpression

        Args:
            format: The type of the expression that is specified in the expression
                property.
            expression: An expression compliant with the format specified. The
                expression can be any data type and depends on the format given,
                e.g. string or object.
        Returns:
            ProcessingExpression
        """
        c = cls({})
        c.apply(format=format, expression=expression)
        return c

    @property
    def format(self) -> str:
        """
        Get or set the required format property of this ProcessingExpression
        """
        return cast(
            str,
            get_required(
                self.properties.get(FORMAT_PROCESSING_EXPRESSION_PROP),
                self,
                FORMAT_PROCESSING_EXPRESSION_PROP,
            ),
        )

    @format.setter
    def format(self, v: str) -> None:
        self.properties[FORMAT_PROCESSING_EXPRESSION_PROP] = v

    @property
    def expression(self) -> Any:
        """
        Get or set the required expression property of this ProcessingExpression
        """
        return get_required(
            self.properties.get(EXPRESSION_PROCESSING_EXPRESSION_PROP),
            self,
            EXPRESSION_PROCESSING_EXPRESSION_PROP,
        )

    @expression.setter
    def expression(self, v: Any) -> None:
        self.properties[EXPRESSION_PROCESSING_EXPRESSION_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        """
        Returns the dictionary encoding of this ProcessingExpression
        Returns:
            dict[str, Any]
        """
        return self.properties


class ValueScaling:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ValueScaling):
            raise NotImplementedError
        return (
            self.type == other.type
            and self.minimum == other.minimum
            and self.maximum == other.maximum
            and self.mean == other.mean
            and self.stddev == other.stddev
            and self.value == other.value
            and self.format == other.format
            and self.expression == other.expression
        )

    def apply(
        self,
        type: ValueScalingType,
        minimum: int | float | None = None,
        maximum: int | float | None = None,
        mean: int | float | None = None,
        stddev: int | float | None = None,
        value: int | float | None = None,
        format: str | None = None,
        expression: str | None = None,
    ) -> None:
        """
        Creates new ValueScaling object. Depending on the type
        parameter, different parameters are required. Consult STAC:MLM documentation
        or use :meth:`~get_required_props` for details on what parameters are required
        for which ValueScaling ``type``.

        Args:
            type: The type of ValueScaling object.
            minimum: A minimum value
            maximum: A maximum value
            mean: A mean value
            stddev: A standard deviation value
            value: A scalar value
            format: The format of the expression
            expression: The expression itself
        """
        self.validate_property_dict(type, locals())

        self.type = type
        self.minimum = minimum
        self.maximum = maximum
        self.mean = mean
        self.stddev = stddev
        self.value = value
        self.format = format
        self.expression = expression

    @classmethod
    def create(
        cls,
        type: ValueScalingType,
        minimum: int | float | None = None,
        maximum: int | float | None = None,
        mean: int | float | None = None,
        stddev: int | float | None = None,
        value: int | float | None = None,
        format: str | None = None,
        expression: str | None = None,
    ) -> ValueScaling:
        """
        Creates new ValueScaling object. Depending on the type
        parameter, different parameters are required. Consult STAC:MLM documentation
        or use :meth:`~get_required_props` for details on what parameters are required
        for which ValueScaling ``type``.

        Args:
            type: The type of ValueScaling object.
            minimum: A minimum value
            maximum: A maximum value
            mean: A mean value
            stddev: A standard deviation value
            value: A scalar value
            format: The format of the expression
            expression: The expression itself

        Returns:
            ValueScaling
        """
        c = cls({})
        c.apply(
            type=type,
            minimum=minimum,
            maximum=maximum,
            mean=mean,
            stddev=stddev,
            value=value,
            format=format,
            expression=expression,
        )
        return c

    @classmethod
    def get_required_props(cls, type: ValueScalingType) -> list[str]:
        """
        Determines the parameters required for a certain ValueScaling operation.

        Args:
            type: The type of ValueScaling operation for which required properties are
                to be retrieved

        Returns:
            list[str]: names of properties required for the given ``type``
        """
        d: dict[str, list[str]] = {
            "min-max": ["minimum", "maximum"],
            "z-score": ["mean", "stddev"],
            "clip": ["minimum", "maximum"],
            "clip-min": ["minimum"],
            "clip-max": ["maximum"],
            "offset": ["value"],
            "scale": ["value"],
            "processing": ["format", "expression"],
        }
        return d[type]

    @classmethod
    def validate_property_dict(
        cls, type: ValueScalingType, props: dict[str, Any]
    ) -> None:
        """
        Validate whether given properties satisfy the requirements set by the
            ValueScaling ``type`` parameter

        Args:
            type: The type of ValueScaling operation
            props: The properties to validate. Keys in this dict are the property
                names, values are the property values

        Raises:
            :class:``STACError``: if the given properties do not satisfy
                the requirements of the ValueScaling ``type``

        """
        required_props = cls.get_required_props(type)
        given_props = [
            prop_name
            for prop_name, prop_value in props.items()
            if prop_value is not None
        ]
        given_props_cleaned = [
            prop for prop in given_props if prop != "self" and prop != "type"
        ]

        valid = all(
            [required_prop in given_props_cleaned for required_prop in required_props]
        )

        if not valid:
            raise STACError(
                f"ValueScaling object of {type=} "
                f"requires properties: {required_props}. Given: {given_props_cleaned}"
            )

    @property
    def type(self) -> str:
        """
        Get or set the required type property of this ValueScaling object
        """
        return cast(
            str,
            get_required(
                self.properties.get(TYPE_VALUE_SCALING_PROP),
                self,
                TYPE_VALUE_SCALING_PROP,
            ),
        )

    @type.setter
    def type(self, v: str) -> None:
        self.properties[TYPE_VALUE_SCALING_PROP] = v

    @property
    def minimum(self) -> int | float | None:
        """
        Get or set the minimum property of this ValueScaling object
        """
        return self.properties.get(MINIMUM_VALUE_SCALING_PROP)

    @minimum.setter
    def minimum(self, v: int | float | None) -> None:
        if v is not None:
            self.properties[MINIMUM_VALUE_SCALING_PROP] = v
        else:
            self.properties.pop(MINIMUM_VALUE_SCALING_PROP, None)

    @property
    def maximum(self) -> int | float | None:
        """
        Get or set the maximum property of this ValueScaling object
        """
        return self.properties.get(MAXIMUM_VALUE_SCALING_PROP)

    @maximum.setter
    def maximum(self, v: int | float | None) -> None:
        if v is not None:
            self.properties[MAXIMUM_VALUE_SCALING_PROP] = v
        else:
            self.properties.get(MAXIMUM_VALUE_SCALING_PROP, None)

    @property
    def mean(self) -> int | float | None:
        """
        Get or set the mean property of this ValueScaling object
        """
        return self.properties.get(MEAN_VALUE_SCALING_PROP)

    @mean.setter
    def mean(self, v: int | float | None) -> None:
        if v is not None:
            self.properties[MEAN_VALUE_SCALING_PROP] = v
        else:
            self.properties.pop(MEAN_VALUE_SCALING_PROP, None)

    @property
    def stddev(self) -> int | float | None:
        """
        Get or set the stddev (standard deviation) property of this ValueScaling object
        """
        return self.properties.get(STDDEV_VALUE_SCALING_PROP)

    @stddev.setter
    def stddev(self, v: int | float | None) -> None:
        if v is not None:
            self.properties[STDDEV_VALUE_SCALING_PROP] = v
        else:
            self.properties.pop(STDDEV_VALUE_SCALING_PROP, None)

    @property
    def value(self) -> int | float | None:
        """
        Get or set the value property of this ValueScaling object
        """
        return self.properties.get(VALUE_VALUE_SCALING_PROP)

    @value.setter
    def value(self, v: int | float | None) -> None:
        if v is not None:
            self.properties[VALUE_VALUE_SCALING_PROP] = v
        else:
            self.properties.pop(VALUE_VALUE_SCALING_PROP, None)

    @property
    def format(self) -> str | None:
        """
        Get or set the format property of this ValueScaling object
        """
        return self.properties.get(FORMAT_VALUE_SCALING_PROP)

    @format.setter
    def format(self, v: str | None) -> None:
        if v is not None:
            self.properties[FORMAT_VALUE_SCALING_PROP] = v
        else:
            self.properties.pop(FORMAT_VALUE_SCALING_PROP, None)

    @property
    def expression(self) -> str | None:
        """
        Get or set the expression property of this ValueScaling object
        """
        return self.properties.get(EXPRESSION_VALUE_SCALING_PROP)

    @expression.setter
    def expression(self, v: str | None) -> None:
        if v is not None:
            self.properties[EXPRESSION_VALUE_SCALING_PROP] = v
        else:
            self.properties.pop(EXPRESSION_VALUE_SCALING_PROP, None)

    def to_dict(self) -> dict[str, Any]:
        """
        Serialize a dict representation of this ValueScaling object

        Returns:
            dict[str, Any]
        """
        return self.properties


class InputStructure:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InputStructure):
            raise NotImplementedError
        return (
            self.shape == other.shape
            and self.dim_order == other.dim_order
            and self.data_type == other.data_type
        )

    def apply(
        self,
        shape: list[int],
        dim_order: list[str],
        data_type: DataType,
    ) -> None:
        """
        Set the properties for a new InputStructure.

        Args:
            shape: Shape of the input n-dimensional array (e.g.: B×C×H×W), including
                the batch size dimension.
                Each dimension must either be greater than 0 or -1 to indicate a
                variable dimension size.
            dim_order: Order of the shape dimensions by name.
            data_type:The data type of values in the n-dimensional array. For model
                inputs, this should be the data type of the processed input supplied to
                the model inference function, not the data type of the source bands.
        """
        self.shape = shape
        self.dim_order = dim_order
        self.data_type = data_type

    @classmethod
    def create(
        cls, shape: list[int], dim_order: list[str], data_type: DataType
    ) -> InputStructure:
        """
        Create a new InputStructure.

        Args:
            shape: Shape of the input n-dimensional array (e.g.: B×C×H×W), including the
                batch size dimension. Each dimension must either be greater than 0 or
                -1 to indicate a variable dimension size.
            dim_order: Order of the shape dimensions by name.
            dim_order: Order of the shape dimensions by name.
            data_type: The data type of values in the n-dimensional array. For model
                inputs, this should be the data type of the processed input supplied to
                the model inference function, not the data type of the source bands.
        Returns:
            InputStructure
        """
        c = cls({})
        c.apply(shape=shape, dim_order=dim_order, data_type=data_type)
        return c

    @property
    def shape(self) -> list[int]:
        """
        Get or set the required shape property of this InputStructure object
        """
        return get_required(
            self.properties.get(SHAPE_INPUT_STRUCTURE_PROP),
            self,
            SHAPE_INPUT_STRUCTURE_PROP,
        )

    @shape.setter
    def shape(self, v: list[int]) -> None:
        self.properties[SHAPE_INPUT_STRUCTURE_PROP] = v

    @property
    def dim_order(self) -> list[str]:
        """
        Get or set the required dim_order property of this InputStructure object
        """
        return get_required(
            self.properties.get(DIM_ORDER_INPUT_STRUCTURE_PROP),
            self,
            DIM_ORDER_INPUT_STRUCTURE_PROP,
        )

    @dim_order.setter
    def dim_order(self, v: list[str]) -> None:
        self.properties[DIM_ORDER_INPUT_STRUCTURE_PROP] = v

    @property
    def data_type(self) -> DataType:
        """
        Get or set the required data_type property of this InputStructure object
        """
        return cast(
            DataType,
            get_required(
                self.properties.get(DATA_TYPE_INPUT_STRUCTURE_PROP),
                self,
                DATA_TYPE_INPUT_STRUCTURE_PROP,
            ),
        )

    @data_type.setter
    def data_type(self, v: DataType) -> None:
        self.properties[DATA_TYPE_INPUT_STRUCTURE_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        """
        Serializes a dict representation of this InputStucture object

        Returns:
            dict[str, Any]
        """
        return self.properties


class ModelInput:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelInput):
            raise NotImplementedError
        return (
            self.name == other.name
            and self.bands == other.bands
            and self.input == other.input
            and self.description == other.description
            and self.value_scaling == other.value_scaling
            and self.resize_type == other.resize_type
            and self.pre_processing_function == other.pre_processing_function
        )

    def apply(
        self,
        name: str,
        bands: list[ModelBand] | list[str],
        input: InputStructure,
        description: str | None = None,
        value_scaling: ValueScaling | None = None,
        resize_type: ResizeType | None = None,
        pre_processing_function: ProcessingExpression | None = None,
    ) -> None:
        """
        Sets the Properties for a new Input

        Args:
            name: Name of the input variable defined by the model. If no explicit name
                is defined by the model, an informative name (e.g.: "RGB Time Series")
                can be used instead.
            bands: The raster band references used to train, fine-tune or perform
                inference with the model, which may be all or a subset of bands
                available in a STAC Item's Band Object. If no band applies for one
                input, use an empty array.
            input: The N-dimensional array definition that describes the shape,
                dimension ordering, and data type.
            description: Additional details about the input such as describing its
                purpose or expected source that cannot be represented by other
                properties.
            value_scaling:Method to scale, normalize, or standardize the data inputs
                values, across dimensions, per corresponding dimension index, or null
                if none applies. These values often correspond to dataset or sensor
                statistics employed for training the model, but can come from another
                source as needed by the model definition. Consider using
                pre_processing_function for custom implementations or more complex
                combinations.
            resize_type: High-level descriptor of the resize method to modify the input
                dimensions shape. Select an appropriate option or null when none
                applies. Consider using pre_processing_function for custom
                implementations or more complex combinations.
            pre_processing_function: Custom preprocessing function where rescaling and
                resize, and any other significant data preparation operations takes
                place. The pre_processing_function should be applied over all available
                bands.
        """
        self.name = name
        self.bands = bands
        self.input = input
        self.description = description
        self.value_scaling = value_scaling
        self.resize_type = resize_type
        self.pre_processing_function = pre_processing_function

    @classmethod
    def create(
        cls,
        name: str,
        bands: list[ModelBand] | list[str],
        input: InputStructure,
        description: str | None = None,
        value_scaling: ValueScaling | None = None,
        resize_type: ResizeType | None = None,
        pre_processing_function: ProcessingExpression | None = None,
    ) -> ModelInput:
        """
        Creates a new Input

        Args:
            name: Name of the input variable defined by the model. If no explicit name
                is defined by the model, an informative name (e.g.: "RGB Time Series")
                can be used instead.
            bands: The raster band references used to train, fine-tune or perform
                inference with the model, which may be all or a subset of bands
                available in a STAC Item's Band Object. If no band applies for one
                input, use an empty array.
            input: The N-dimensional array definition that describes the shape,
                dimension ordering, and data type.
            description: Additional details about the input such as describing its
                purpose or expected source that cannot be represented by other
                properties.
            value_scaling:Method to scale, normalize, or standardize the data inputs
                values, across dimensions, per corresponding dimension index, or null
                if none applies. These values often correspond to dataset or sensor
                statistics employed for training the model, but can come from another
                source as needed by the model definition. Consider using
                pre_processing_function for custom implementations or more complex
                combinations.
            resize_type: High-level descriptor of the resize method to modify the input
                dimensions shape. Select an appropriate option or null when none
                applies. Consider using pre_processing_function for custom
                implementations or more complex combinations.
            pre_processing_function: Custom preprocessing function where rescaling and
                resize, and any other significant data preparation operations takes
                place. The pre_processing_function should be applied over all available
                bands.
        """
        c = cls({})
        c.apply(
            name=name,
            bands=bands,
            input=input,
            description=description,
            value_scaling=value_scaling,
            resize_type=resize_type,
            pre_processing_function=pre_processing_function,
        )
        return c

    @property
    def name(self) -> str:
        """
        Gets or sets the required name property of this ModelInput object
        """
        return cast(
            str,
            get_required(
                self.properties.get(NAME_INPUT_OBJECT_PROP),
                self,
                NAME_INPUT_OBJECT_PROP,
            ),
        )

    @name.setter
    def name(self, v: str) -> None:
        self.properties[NAME_INPUT_OBJECT_PROP] = v

    @property
    def bands(self) -> list[ModelBand] | list[str]:
        """
        Gets or sets the required bands property of this ModelInput object
        """
        bands: list[str] | list[dict[str, Any]] = get_required(
            self.properties.get(BANDS_INPUT_OBJECT_PROP), self, BANDS_INPUT_OBJECT_PROP
        )

        if isinstance(bands, list) and all(isinstance(item, str) for item in bands):
            return [band for band in bands if isinstance(band, str)]

        elif isinstance(bands, list) and all(isinstance(item, dict) for item in bands):
            return [ModelBand(band) for band in bands if isinstance(band, dict)]

        raise TypeError("Invalid bands property. Must list[str] or list[ModelBand]")

    @bands.setter
    def bands(self, v: list[ModelBand] | list[str]) -> None:
        v_trans = [c.to_dict() if isinstance(c, ModelBand) else c for c in v]
        self.properties[BANDS_INPUT_OBJECT_PROP] = v_trans

    @property
    def input(self) -> InputStructure:
        """
        Gets or sets the required input property of this ModelInput object
        """
        return InputStructure(
            get_required(
                self.properties.get(INPUT_INPUT_OBJECT_PROP),
                self,
                INPUT_INPUT_OBJECT_PROP,
            )
        )

    @input.setter
    def input(self, v: InputStructure) -> None:
        self.properties[INPUT_INPUT_OBJECT_PROP] = v.to_dict()

    @property
    def description(self) -> str | None:
        """
        Gets or sets the description property of this ModelInput object
        """
        return self.properties.get(DESCRIPTION_INPUT_OBJECT_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is not None:
            self.properties[DESCRIPTION_INPUT_OBJECT_PROP] = v
        else:
            self.properties.pop(DESCRIPTION_INPUT_OBJECT_PROP, None)

    @property
    def value_scaling(self) -> ValueScaling | None:
        """
        Gets or sets the value_scaling property of this ModelInput object
        """
        v = self.properties.get(VALUE_SCALING_INPUT_OBJECT_PROP)
        return ValueScaling(v) if v is not None else v

    @value_scaling.setter
    def value_scaling(self, v: ValueScaling | None) -> None:
        # add None to properties dict and do not pop it, according to specification
        self.properties[VALUE_SCALING_INPUT_OBJECT_PROP] = (
            None if v is None else v.to_dict()
        )

    @property
    def resize_type(self) -> ResizeType | None:
        """
        Gets or sets the resize_type property of this ModelInput object
        """
        return self.properties.get(RESIZE_TYPE_INPUT_OBJECT_PROP)

    @resize_type.setter
    def resize_type(self, v: ResizeType | None) -> None:
        # add to dict even if v is None and do not pop it, according to specification
        self.properties[RESIZE_TYPE_INPUT_OBJECT_PROP] = v

    @property
    def pre_processing_function(self) -> ProcessingExpression | None:
        """
        Gets or sets the pre_processing_function property of this ModelInput object
        """
        v = self.properties.get(PRE_PROCESSING_FUNCTION_INPUT_OBJECT_PROP)
        return ProcessingExpression(v) if v is not None else None

    @pre_processing_function.setter
    def pre_processing_function(self, v: ProcessingExpression | None) -> None:
        # add to dict even if v is None and do not pop it, according to specification
        self.properties[PRE_PROCESSING_FUNCTION_INPUT_OBJECT_PROP] = (
            None if v is None else v.to_dict()
        )

    def to_dict(self) -> dict[str, Any]:
        """
        Serializes this ModelInput object into its dict representation

        Returns:
            dict[str, Any]
        """
        return self.properties


class ResultStructure:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ResultStructure):
            raise NotImplementedError
        return (
            self.shape == other.shape
            and self.dim_order == other.dim_order
            and self.data_type == other.data_type
        )

    def apply(
        self, shape: list[int], dim_order: list[str], data_type: DataType
    ) -> None:
        """
        Set the properties for a new ResultStructure.

        Args:
            shape: Shape of the n-dimensional result array (e.g.: B×H×W or B×C),
                possibly including a batch size dimension. The dimensions must either
                be greater than 0 or -1 to indicate a variable size.
            dim_order: Order of the shape dimensions by name for the result array.
            data_type: The data type of values in the n-dimensional array. For model
                outputs, this should be the data type of the result of the model
                inference without extra post-processing.
        """
        self.shape = shape
        self.dim_order = dim_order
        self.data_type = data_type

    @classmethod
    def create(
        cls, shape: list[int], dim_order: list[str], data_type: DataType
    ) -> ResultStructure:
        """
        Creates a new ResultStructure.

        Args:
            shape: Shape of the n-dimensional result array (e.g.: B×H×W or B×C),
                possibly including a batch size dimension. The dimensions must either
                be greater than 0 or -1 to indicate a variable size.
            dim_order: Order of the shape dimensions by name for the result array.
            data_type: The data type of values in the n-dimensional array. For model
                outputs, this should be the data type of the result of the model
                inference without extra post-processing.

        Returns:
            ResultStructure
        """
        c = cls({})
        c.apply(shape=shape, dim_order=dim_order, data_type=data_type)
        return c

    @property
    def shape(self) -> list[int]:
        """
        Gets or sets the required shape property of the ResultStructure object
        """
        return get_required(
            self.properties.get(SHAPE_RESULT_STRUCTURE_PROP),
            self,
            SHAPE_RESULT_STRUCTURE_PROP,
        )

    @shape.setter
    def shape(self, v: list[int]) -> None:
        self.properties[SHAPE_RESULT_STRUCTURE_PROP] = v

    @property
    def dim_order(self) -> list[str]:
        """
        Gets or sets the required dim_order property of the ResultStructure object
        """
        return get_required(
            self.properties.get(DIM_ORDER_RESULT_STRUCTURE_PROP),
            self,
            DIM_ORDER_RESULT_STRUCTURE_PROP,
        )

    @dim_order.setter
    def dim_order(self, v: list[str]) -> None:
        self.properties[DIM_ORDER_RESULT_STRUCTURE_PROP] = v

    @property
    def data_type(self) -> DataType:
        """
        Gets or sets the required data_type property of the ResultStructure object
        """
        return cast(
            DataType,
            get_required(
                self.properties.get(DATA_TYPE_RESULT_STRUCTURE_PROP),
                self,
                DIM_ORDER_RESULT_STRUCTURE_PROP,
            ),
        )

    @data_type.setter
    def data_type(self, v: DataType) -> None:
        self.properties[DATA_TYPE_RESULT_STRUCTURE_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        """
        Serilaizes this ResultStructure object to a dict
        Returns:
            dict[str, Any]
        """
        return self.properties


class ModelOutput:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelOutput):
            raise NotImplementedError
        return (
            self.name == other.name
            and self.tasks == other.tasks
            and self.result == other.result
            and self.description == other.description
            and self.classes == other.classes
            and self.post_processing_function == other.post_processing_function
        )

    def apply(
        self,
        name: str,
        tasks: list[TaskType],
        result: ResultStructure,
        description: str | None = None,
        classes: list[Classification] | None = None,
        post_processing_function: ProcessingExpression | None = None,
    ) -> None:
        """
        Sets the properties for a new Output

        Args:
            name:Name of the output variable defined by the model. If no explicit name
                is defined by the model, an informative name (e.g.: "CLASSIFICATION")
                can be used instead.
            tasks: Specifies the Machine Learning tasks for which the output can be used
                for. This can be a subset of mlm:tasks defined under the Item properties
                as applicable.
            result: The structure that describes the resulting output arrays/tensors
                from one model head.
            description: Additional details about the output such as describing its
                purpose or expected result that cannot be represented by other
                properties.
            classes: A list of class objects adhering to the Classification Extension.
            post_processing_function: Custom postprocessing function where
                normalization, rescaling, or any other
                significant operations takes place.
        """
        self.name = name
        self.tasks = tasks
        self.result = result
        self.description = description
        self.classes = classes
        self.post_processing_function = post_processing_function

    @classmethod
    def create(
        cls,
        name: str,
        tasks: list[TaskType],
        result: ResultStructure,
        description: str | None = None,
        classes: list[Classification] | None = None,
        post_processing_function: ProcessingExpression | None = None,
    ) -> ModelOutput:
        """
        Creates a new Output

        Args:
            name:Name of the output variable defined by the model. If no explicit name
                is defined by the model, an informative name (e.g.: "CLASSIFICATION")
                can be used instead.
            tasks: Specifies the Machine Learning tasks for which the output can be used
                for. This can be a subset of mlm:tasks defined under the Item properties
                as applicable.
            result: The structure that describes the resulting output arrays/tensors
                from one model head. description: Additional details about the output
                such as describing its purpose or expected result that cannot be
                represented by other properties.
            classes: A list of class objects adhering to the Classification Extension.
            post_processing_function: Custom postprocessing function where
                normalization, rescaling, or any other significant operations takes
                place.

        Returns:
            ModelOutput
        """
        c = cls({})
        c.apply(
            name=name,
            tasks=tasks,
            result=result,
            description=description,
            classes=classes,
            post_processing_function=post_processing_function,
        )
        return c

    @property
    def name(self) -> str:
        """
        Gets or sets the required name property of this ModelOutput object
        """
        return cast(
            str,
            get_required(self.properties.get(NAME_RESULT_PROP), self, NAME_RESULT_PROP),
        )

    @name.setter
    def name(self, v: str) -> None:
        self.properties[NAME_RESULT_PROP] = v

    @property
    def tasks(self) -> list[TaskType]:
        """
        Gets or sets the required tasks property of this ModelOutput object
        """
        return get_required(
            self.properties.get(TASKS_RESULT_PROP), self, TASKS_RESULT_PROP
        )

    @tasks.setter
    def tasks(self, v: list[TaskType]) -> None:
        self.properties[TASKS_RESULT_PROP] = v

    @property
    def result(self) -> ResultStructure:
        """
        Gets or sets the required results property of this ModelOutput object
        """
        return ResultStructure(
            get_required(
                self.properties.get(RESULT_RESULT_PROP), self, RESULT_RESULT_PROP
            )
        )

    @result.setter
    def result(self, v: ResultStructure) -> None:
        self.properties[RESULT_RESULT_PROP] = v.to_dict()

    @property
    def description(self) -> str | None:
        """
        Gets or sets the description property of this ModelOutput object
        """
        return self.properties.get(DESCRIPTION_RESULT_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is not None:
            self.properties[DESCRIPTION_RESULT_PROP] = v
        else:
            self.properties.pop(DESCRIPTION_RESULT_PROP, None)

    @property
    def classes(self) -> list[Classification] | None:
        """
        Gets or sets the classes property of this ModelOutput object
        """
        classes = self.properties.get(CLASSES_RESULT_PROP)
        return [Classification(c) for c in classes] if classes is not None else None

    @classes.setter
    def classes(self, v: list[Classification] | None) -> None:
        if v is not None:
            self.properties[CLASSES_RESULT_PROP] = [c.to_dict() for c in v]
        else:
            self.properties.pop(CLASSES_RESULT_PROP, None)

    @property
    def post_processing_function(self) -> ProcessingExpression | None:
        """
        Gets or sets the post_processing_function property of this ModelOutput object
        """
        v = self.properties.get(POST_PROCESSING_FUNCTION_RESULT_PROP)
        return ProcessingExpression(v) if v is not None else None

    @post_processing_function.setter
    def post_processing_function(self, v: ProcessingExpression | None) -> None:
        if v is not None:
            self.properties[POST_PROCESSING_FUNCTION_RESULT_PROP] = v.to_dict()
        else:
            self.properties.pop(POST_PROCESSING_FUNCTION_RESULT_PROP, None)

    def to_dict(self) -> dict[str, Any]:
        """
        Serializes this ModelOutput object into a dict
        Returns:
            dict[str, Any]
        """
        return self.properties


class Hyperparameters:
    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Hyperparameters):
            raise NotImplementedError
        return self.properties == other.properties

    def apply(self, **kwargs: Any) -> None:
        """
        Sets the properties for a new Hyperparameters object. The stac:mlm specification
        defines neither their names nor their values, so any key-value pair is allowed.
        Args:
            **kwargs: any model hyperparameter name and its value, as key-value paris
        """
        self.properties.update(kwargs)

    @classmethod
    def create(cls, **kwargs: Any) -> Hyperparameters:
        c = cls({})
        c.apply(**kwargs)
        return c

    def to_dict(self) -> dict[str, Any]:
        """
        Serializes this Hyperparameters object into a dictionary
        Returns:
            dict[str, Any]
        """
        return self.properties


class MLMExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend to properties of an
    :class:`pystac.Item` or :class:`pystac.Collection` with properties from the
    :stac-ext:`Machine Learning Model Extension <mlm>`.

    This class can be used to extend :class:`pystac.Item`, :class:`pystac.Collection`
    and :class:`pystac.ItemAssetDefinition`. For extending :class:`pystac.Asset`, use
    either :class:`~AssetGeneralMLMExtension`: or :class:`AssetDetailedMLMExtension`.
    """

    name: Literal["mlm"] = "mlm"
    properties: dict[str, Any]

    def apply(
        self,
        name: str,
        architecture: str,
        tasks: list[TaskType],
        input: list[ModelInput],
        output: list[ModelOutput],
        framework: str | None = None,
        framework_version: str | None = None,
        memory_size: int | None = None,
        total_parameters: int | None = None,
        pretrained: bool | None = None,
        pretrained_source: str | None = None,
        batch_size_suggestion: int | None = None,
        accelerator: AcceleratorType | None = None,
        accelerator_constrained: bool | None = None,
        accelerator_summary: str | None = None,
        accelerator_count: int | None = None,
        hyperparameters: Hyperparameters | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Sets the properties of a new MLMExtension

        Args:
            name:  name for the model
            architecture: A generic and well established architecture name of the model
            tasks: Specifies the Machine Learning tasks for which the model can be
                used for
            input: Describes the transformation between the EO data and the model input
            output: Describes each model output and how to interpret it.
            framework: Framework used to train the model
            framework_version: The ``framework`` library version
            memory_size: The in-memory size of the model on the accelerator during
                inference (bytes)
            total_parameters: Total number of model parameters, including trainable and
                non-trainable parameters.
            pretrained: Indicates if the model was pretrained. If the model was
                pretrained, consider providing ``pretrained_source`` if it is known
            pretrained_source: The source of the pretraining.
            batch_size_suggestion: A suggested batch size for the accelerator and
                summarized hardware.
            accelerator: The intended computational hardware that runs inference
            accelerator_constrained: Indicates if the intended ``accelerator`` is the
                only accelerator that can run inference
            accelerator_summary: A high level description of the ``accelerator``
            accelerator_count: A minimum amount of ``accelerator`` instances required to
                run the model
            hyperparameters: Additional hyperparameters relevant for the model
            *args: Unused (no effect, only here for signature compliance with apply
                method in derived classes
            **kwargs: Unused (no effect, only here for signature compliance with apply
                method in derived classes
        """
        self.mlm_name = name
        self.architecture = architecture
        self.tasks = tasks
        self.input = input
        self.output = output
        self.framework = framework
        self.framework_version = framework_version
        self.memory_size = memory_size
        self.total_parameters = total_parameters
        self.pretrained = pretrained
        self.pretrained_source = pretrained_source
        self.batch_size_suggestion = batch_size_suggestion
        self.accelerator = accelerator
        self.accelerator_constrained = accelerator_constrained
        self.accelerator_summary = accelerator_summary
        self.accelerator_count = accelerator_count
        self.hyperparameters = hyperparameters

    @classmethod
    def get_schema_uri(cls) -> str:
        """
        Retrieves this extension's schema URI

        Returns:
            str: the schema URI
        """
        return SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> MLMExtension[T]:
        """
        Extend a STAC object (``obj``) with the MLMExtension

        Args:
            obj: The STAC object to be extended.
            add_if_missing: Defines whether this extension's URI should be added to
                this object's  (or its parent's) list of extensions if it is not already
                listed there.

        Returns:
            MLMExtension[T]: The extended object

        Raises:
            TypeError: When a :class:`pystac.Asset` object is passed as the
                `obj` parameter
            pystac.ExtensionTypeError: When any unsupported object is passed as the
                `obj` parameter. If you see this extension in this context, please
                raise an issue on GitHub.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(MLMExtension[T], ItemMLMExtension(obj))
        elif isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(MLMExtension[T], CollectionMLMExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(MLMExtension[T], ItemAssetMLMExtension(obj))
        elif isinstance(obj, pystac.Asset):
            raise TypeError(
                "This class cannot be used to extend STAC objects of type Assets. "
                "To extend Asset objects, use either AssetGeneralMLMExtension or "
                "AssetDetailedMLMExtension"
            )
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @property
    def mlm_name(self) -> str:
        """
        Get or set the required (mlm) name property. It is named mlm_name in this
        context to not break convention and overwrite the extension name class property.
        """
        return cast(str, get_required(self.properties.get(NAME_PROP), self, NAME_PROP))

    @mlm_name.setter
    def mlm_name(self, v: str) -> None:
        self._set_property(NAME_PROP, v)

    @property
    def architecture(self) -> str:
        """
        Get or set the required architecture property
        """
        return cast(
            str,
            get_required(
                self.properties.get(ARCHITECTURE_PROP), self, ARCHITECTURE_PROP
            ),
        )

    @architecture.setter
    def architecture(self, v: str) -> None:
        self._set_property(ARCHITECTURE_PROP, v)

    @property
    def tasks(self) -> list[TaskType]:
        """
        Get or set the required tasks property
        """
        return get_required(self.properties.get(TASKS_PROP), self, TASKS_PROP)

    @tasks.setter
    def tasks(self, v: list[TaskType]) -> None:
        self._set_property(TASKS_PROP, v)

    @property
    def framework(self) -> str | None:
        """
        Get or set the framework property
        """
        return self._get_property(FRAMEWORK_PROP, str)

    @framework.setter
    def framework(self, v: str | None) -> None:
        self._set_property(FRAMEWORK_PROP, v)

    @property
    def framework_version(self) -> str | None:
        """
        Get or set the framework_version property
        """
        return self._get_property(FRAMEWORK_VERSION_PROP, str)

    @framework_version.setter
    def framework_version(self, v: str | None) -> None:
        self._set_property(FRAMEWORK_VERSION_PROP, v)

    @property
    def memory_size(self) -> int | None:
        """
        Get or set the memory_size property
        """
        return self._get_property(MEMORY_SIZE_PROP, int)

    @memory_size.setter
    def memory_size(self, v: int | None) -> None:
        self._set_property(MEMORY_SIZE_PROP, v)

    @property
    def total_parameters(self) -> int | None:
        """
        Get or set the total_parameters property
        """
        return self._get_property(TOTAL_PARAMETERS_PROP, int)

    @total_parameters.setter
    def total_parameters(self, v: int | None) -> None:
        self._set_property(TOTAL_PARAMETERS_PROP, v)

    @property
    def pretrained(self) -> bool | None:
        """
        Get or set the pretrained property
        """
        return self._get_property(PRETRAINED_PROP, bool)

    @pretrained.setter
    def pretrained(self, v: bool | None) -> None:
        self._set_property(PRETRAINED_PROP, v)

    @property
    def pretrained_source(self) -> str | None:
        """
        Get or set the pretrained_source property
        """
        return self._get_property(PRETRAINED_SOURCE_PROP, str)

    @pretrained_source.setter
    def pretrained_source(self, v: str | None) -> None:
        self._set_property(
            PRETRAINED_SOURCE_PROP, v, False
        )  # dont pop as per documentation

    @property
    def batch_size_suggestion(self) -> int | None:
        """
        Get or set the batch_size_suggestion property
        """
        return self._get_property(BATCH_SIZE_SUGGESTION_PROP, int)

    @batch_size_suggestion.setter
    def batch_size_suggestion(self, v: int | None) -> None:
        self._set_property(BATCH_SIZE_SUGGESTION_PROP, v)

    @property
    def accelerator(self) -> AcceleratorType | None:
        """
        Get or set the accelerator property
        """
        return self._get_property(ACCELERATOR_PROP, AcceleratorType)

    @accelerator.setter
    def accelerator(self, v: AcceleratorType | None) -> None:
        # dont pop as per documentation
        self._set_property(ACCELERATOR_PROP, v, False)

    @property
    def accelerator_constrained(self) -> bool | None:
        """
        Get or set the accelerator_constrained property
        """
        return self._get_property(ACCELERATOR_CONSTRAINED_PROP, bool)

    @accelerator_constrained.setter
    def accelerator_constrained(self, v: bool | None) -> None:
        self._set_property(ACCELERATOR_CONSTRAINED_PROP, v)

    @property
    def accelerator_summary(self) -> str | None:
        """
        Get or set the accelerator_summary property
        """
        return self._get_property(ACCELERATOR_SUMMARY_PROP, str)

    @accelerator_summary.setter
    def accelerator_summary(self, v: str | None) -> None:
        self._set_property(ACCELERATOR_SUMMARY_PROP, v)

    @property
    def accelerator_count(self) -> int | None:
        """
        Get or set the accelerator_count property
        """
        return self._get_property(ACCELERATOR_COUNT_PROP, int)

    @accelerator_count.setter
    def accelerator_count(self, v: int | None) -> None:
        self._set_property(ACCELERATOR_COUNT_PROP, v)

    @property
    def input(self) -> list[ModelInput]:
        """
        Get or set the required input property
        """
        return [
            ModelInput(inp)
            for inp in get_required(
                self._get_property(INPUT_PROP, list[dict[str, Any]]), self, INPUT_PROP
            )
        ]

    @input.setter
    def input(self, v: list[ModelInput]) -> None:
        self._set_property(INPUT_PROP, [inp.to_dict() for inp in v])

    @property
    def output(self) -> list[ModelOutput]:
        """
        Get or set the required output property
        """
        return [
            ModelOutput(outp)
            for outp in get_required(
                self._get_property(OUTPUT_PROP, list[dict[str, Any]]), self, OUTPUT_PROP
            )
        ]

    @output.setter
    def output(self, v: list[ModelOutput]) -> None:
        self._set_property(OUTPUT_PROP, [outp.to_dict() for outp in v])

    @property
    def hyperparameters(self) -> Hyperparameters | None:
        """
        Get or set the hyperparameters property
        """
        prop = self._get_property(HYPERPARAMETERS_PROP, dict[str, Any])
        return Hyperparameters(prop) if prop is not None else None

    @hyperparameters.setter
    def hyperparameters(self, v: Hyperparameters | None) -> None:
        self._set_property(HYPERPARAMETERS_PROP, v.to_dict() if v is not None else None)

    def to_dict(self) -> dict[str, Any]:
        """
        Serializes this MLMExtension object into a dict

        Returns:
            dict[str, Any]
        """
        return self.properties


class ItemMLMExtension(MLMExtension[pystac.Item]):
    properties: dict[str, Any]
    item: pystac.Item

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemMLMExtension Item id={self.item.id}>"


class CollectionMLMExtension(MLMExtension[pystac.Collection]):
    collection: pystac.Collection
    properties: dict[str, Any]
    links: list[pystac.Link]

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields
        self.links = collection.links

    def __repr__(self) -> str:
        return f"<CollectionMLMExtension Collection id={self.collection.id}>"


class _AssetMLMExtension(ABC):
    """
    Abstract base class for (derived) MLM asset extensions.
    """

    name: Literal["mlm"] = "mlm"
    asset: pystac.Asset
    asset_href: str
    properties: dict[str, Any]
    additional_read_properties: Iterable[dict[str, Any]] | None

    def __init__(self, asset: pystac.Asset):
        self.asset = asset
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    @classmethod
    def _ext(cls: type[AssetExtensionType], obj: pystac.Asset) -> AssetExtensionType:
        if not isinstance(obj, pystac.Asset):
            raise TypeError(
                "This class can only be used to extend Assets. "
                "For Items and Collections use MLMExtension."
            )
        return cls(obj)

    def to_dict(self) -> dict[str, Any]:
        """
        Serializes the extended asset's properties into a dict

        Returns:
            dict[str, Any]
        """
        return self.properties

    @classmethod
    def get_schema_uri(cls) -> str:
        """
        Retrieve this extension's schema URI

        Returns:
            str: Schema URI
        """
        return SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)

    @property
    def artifact_type(self) -> str | None:
        """
        Get or set the artifact_type property. This is required if ``mlm:model`` as one
         of the asset's roles
        """
        prop_value = self.properties.get(ARTIFACT_TYPE_ASSET_PROP)
        if isinstance(self.asset.roles, list) and "mlm:model" in self.asset.roles:
            return cast(str, get_required(prop_value, self, ARTIFACT_TYPE_ASSET_PROP))
        else:
            return prop_value

    @artifact_type.setter
    def artifact_type(self, v: str | None) -> None:
        if isinstance(self.asset.roles, list) and "mlm:model" in self.asset.roles:
            if v is None:
                raise pystac.errors.RequiredPropertyMissing(
                    self,
                    ARTIFACT_TYPE_ASSET_PROP,
                    f"{ARTIFACT_TYPE_ASSET_PROP} is a required property and must "
                    f"not be None for asset with role mlm:model.",
                )
            self.properties[ARTIFACT_TYPE_ASSET_PROP] = v
        else:
            if v is not None:
                self.properties[ARTIFACT_TYPE_ASSET_PROP] = v
            else:
                self.properties.pop(ARTIFACT_TYPE_ASSET_PROP, None)

    @property
    def compile_method(self) -> str | None:
        """
        Get or set this asset's compile_method property
        """
        return self.properties.get(COMPILE_METHOD_ASSET_PROP)

    @compile_method.setter
    def compile_method(self, v: str | None) -> None:
        if v is not None:
            self.properties[COMPILE_METHOD_ASSET_PROP] = v
        else:
            self.properties.pop(COMPILE_METHOD_ASSET_PROP, None)

    @property
    def entrypoint(self) -> str | None:
        """
        Get or set this asset's entrypoint property
        """
        return self.properties.get(ENTRYPOINT_ASSET_PROP)

    @entrypoint.setter
    def entrypoint(self, v: str | None) -> None:
        if v is not None:
            self.properties[ENTRYPOINT_ASSET_PROP] = v
        else:
            self.properties.pop(ENTRYPOINT_ASSET_PROP, None)


class AssetGeneralMLMExtension(
    _AssetMLMExtension,
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """A class that can be used to extend the properties of an
    :class:`pystac.Asset` object with properties from the
    :stac-ext:`Machine Learning Model Extension <mlm>`.

    Use this class, if model metadata is provided by the asset's parent object
    (i.e. :class:`pystac.Item` or :class:`pystac.Item`). If Model metadata is provided
    by the asset object itself, use :class:`AssetDetailedMLMExtension`.

    For extending :class:`pystac.Item`, :class:`pystac.Collection` or
    :class:`pystac.ItemAssetDefinition` objects, use :class:`MLMExtension` instead.
    """

    def apply(
        self,
        artifact_type: str | None = None,
        compile_method: str | None = None,
        entrypoint: str | None = None,
    ) -> None:
        """
        Sets the properties of a new AssetGeneralMLMExtension

        Args:
            artifact_type: Specifies the kind of model artifact, any string is allowed.
                Typically related to a particular ML framework. This property is
                required when ``mlm:model`` is listed as a role of this asset
            compile_method: Describes the method used to compile the ML model either
                when the model is saved or at model runtime prior to inference
            entrypoint: Specific entrypoint reference in the code to use for running
                model inference.
        """
        self.artifact_type = artifact_type
        self.compile_method = compile_method
        self.entrypoint = entrypoint

    @classmethod
    def ext(
        cls, obj: pystac.Asset, add_if_missing: bool = False
    ) -> AssetGeneralMLMExtension[pystac.Asset]:
        """
        Extend a :class:`pystac.Asset` (``obj``) with the AssetGeneralMLMExtension

        Args:
            obj: The Asset to be extended.
            add_if_missing: Defines whether this extension's URI should be added to
                this asset's parent's list of extensions if it is not already
                listed there. Use ``False`` if the asset does not specify a parent

        Returns:
            AssetGeneralMLMExtension[pystac.Asset]: The extended object
        """
        cls.ensure_owner_has_extension(obj, add_if_missing)
        return AssetGeneralMLMExtension._ext(obj)

    def __repr__(self) -> str:
        return f"<AssetGeneralMLMExtension Asset href={self.asset_href}>"


class AssetDetailedMLMExtension(_AssetMLMExtension, MLMExtension[pystac.Asset]):
    """A class that can be used to extend the properties of an
    :class:`pystac.Asset` object with properties from the
    :stac-ext:`Machine Learning Model Extension <mlm>`.

    Use this class, if model metadata is provided by the asset. If model metadata is
    provided by the asset's parent object
    (i.e. :class:`pystac.Item` or :class:`pystac.Item`), use
    :class:`AssetGeneralMLMExtension`.

    For extending :class:`pystac.Item`, :class:`pystac.Collection` or
    :class:`pystac.ItemAssetDefinition` objects, use :class:`MLMExtension` instead.
    """

    def __repr__(self) -> str:
        return f"<AssetDetailedMLMExtension Asset href={self.asset_href}>"

    @classmethod
    def ext(
        cls, obj: pystac.Asset, add_if_missing: bool = False
    ) -> AssetDetailedMLMExtension:
        """
        Extend a :class:`pystac.Asset` (``obj``) with the AssetDetailedMLMExtension

        Args:
            obj: The Asset to be extended.
            add_if_missing: Defines whether this extension's URI should be added to
                this asset's parent's list of extensions if it is not already
                listed there. Use ``False`` if the asset does not specify a parent

        Returns:
            AssetDetailedMLMExtension[pystac.Asset]: The extended object
        """
        cls.ensure_owner_has_extension(obj, add_if_missing)
        return AssetDetailedMLMExtension._ext(obj)

    def apply(
        self,
        name: str,
        architecture: str,
        tasks: list[TaskType],
        input: list[ModelInput],
        output: list[ModelOutput],
        framework: str | None = None,
        framework_version: str | None = None,
        memory_size: int | None = None,
        total_parameters: int | None = None,
        pretrained: bool | None = None,
        pretrained_source: str | None = None,
        batch_size_suggestion: int | None = None,
        accelerator: AcceleratorType | None = None,
        accelerator_constrained: bool | None = None,
        accelerator_summary: str | None = None,
        accelerator_count: int | None = None,
        hyperparameters: Hyperparameters | None = None,
        artifact_type: str | None = None,
        compile_method: str | None = None,
        entrypoint: str | None = None,
    ) -> None:
        """
        Sets the properties of a new AssetDetailedMLMExtensions

        Args:
            name:  name for the model
            architecture: A generic and well established architecture name of the model
            tasks: Specifies the Machine Learning tasks for which the model can be
                used for
            input: Describes the transformation between the EO data and the model input
            output: Describes each model output and how to interpret it.
            framework: Framework used to train the model
            framework_version: The ``framework`` library version
            memory_size: The in-memory size of the model on the accelerator during
                inference (bytes)
            total_parameters: Total number of model parameters, including trainable and
                non-trainable parameters.
            pretrained: Indicates if the model was pretrained. If the model was
                pretrained, consider providing ``pretrained_source`` if it is known
            pretrained_source: The source of the pretraining.
            batch_size_suggestion: A suggested batch size for the accelerator and
                summarized hardware.
            accelerator: The intended computational hardware that runs inference
            accelerator_constrained: Indicates if the intended ``accelerator`` is the
                only accelerator that can run inference
            accelerator_summary: A high level description of the ``accelerator``
            accelerator_count: A minimum amount of ``accelerator`` instances required to
                run the model
            hyperparameters: Additional hyperparameters relevant for the model
            artifact_type: Specifies the kind of model artifact, any string is allowed.
                Typically related to a particular ML framework. This property is
                required when ``mlm:model`` is listed as a role of this asset
            compile_method: Describes the method used to compile the ML model either
                when the model is saved or at model runtime prior to inference
            entrypoint: Specific entrypoint reference in the code to use for running
                model inference.
        """
        MLMExtension.apply(
            self,
            name=name,
            architecture=architecture,
            tasks=tasks,
            input=input,
            output=output,
            framework=framework,
            framework_version=framework_version,
            memory_size=memory_size,
            total_parameters=total_parameters,
            pretrained=pretrained,
            pretrained_source=pretrained_source,
            batch_size_suggestion=batch_size_suggestion,
            accelerator=accelerator,
            accelerator_constrained=accelerator_constrained,
            accelerator_summary=accelerator_summary,
            accelerator_count=accelerator_count,
            hyperparameters=hyperparameters,
        )
        self.artifact_type = artifact_type
        self.compile_method = compile_method
        self.entrypoint = entrypoint


class ItemAssetMLMExtension(MLMExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties

    def __repr__(self) -> str:
        return f"<ItemAssetsMLMExtension ItemAssetDefinition={self.asset_defn}"


class MLMExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)
    prev_extension_ids = {
        SCHEMA_URI_PATTERN.format(version=v)
        for v in SUPPORTED_VERSIONS
        if v != DEFAULT_VERSION
    }
    stac_object_types = {pystac.STACObjectType.ITEM, pystac.STACObjectType.COLLECTION}

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        # mo adjustments to objects needed when migrating yet
        # schema back to v1.0 is fully backwards compatible
        super().migrate(obj, version, info)


MLM_EXTENSION_HOOKS: ExtensionHooks = MLMExtensionHooks()



import json
import os
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, cast

import pytest

import pystac
from pystac import Collection, Item, Link
from pystac.errors import STACError
from pystac.link import HIERARCHICAL_LINKS
from pystac.utils import make_posix_style
from tests.utils.test_cases import ARBITRARY_EXTENT

TEST_DATETIME: datetime = datetime(2020, 3, 14, 16, 32)


def test_path_like() -> None:
    rel = "some-rel"
    target = os.path.abspath("../elsewhere")
    link = pystac.Link(rel, target)
    assert os.fspath(link) == make_posix_style(target)


def test_minimal(item: pystac.Item) -> None:
    rel = "my rel"
    target = "https://example.com/a/b"
    link = pystac.Link(rel, target)
    assert target == link.get_href()
    assert target == link.get_absolute_href()

    expected_repr = f"<Link rel={rel} target={target}>"
    assert expected_repr == link.__repr__()

    assert not link.is_resolved()

    expected_dict = {"rel": rel, "href": target}
    assert expected_dict == link.to_dict()

    # Run the same tests on the clone.
    clone = link.clone()
    assert link != clone

    assert target == clone.get_href()
    assert target == clone.get_absolute_href()

    assert expected_repr == clone.__repr__()

    assert expected_dict == clone.to_dict()

    # Try the modification methods.
    assert link.owner is None
    link.set_owner(None)
    assert link.owner is None

    link.set_owner(item)
    assert item == link.owner


def test_relative() -> None:
    rel = "my rel"
    target = "../elsewhere"
    mime_type = "example/stac_thing"
    link = pystac.Link(rel, target, mime_type, "a title", extra_fields={"a": "b"})
    expected_dict = {
        "rel": rel,
        "href": target,
        "type": "example/stac_thing",
        "title": "a title",
        "a": "b",
    }
    assert expected_dict == link.to_dict()


def test_link_does_not_fail_if_href_is_none(item: pystac.Item) -> None:
    """Test to ensure get_href does not fail when the href is None."""
    catalog = pystac.Catalog(id="test", description="test desc")
    catalog.add_item(item)
    catalog.set_self_href("/some/href")

    link = catalog.get_single_link("item")
    assert link is not None
    assert link.get_href() is None


def test_resolve_stac_object_no_root_and_target_is_item(item: pystac.Item) -> None:
    link = pystac.Link("my rel", target=item)
    link.resolve_stac_object()


@pytest.mark.skipif(os.name == "nt", reason="Non-windows test")
def test_resolve_stac_object_throws_informative_error() -> None:
    link = pystac.Link("root", target="/a/b/foo.json")
    with pytest.raises(
        STACError, match="HREF: '/a/b/foo.json' does not resolve to a STAC object"
    ):
        link.resolve_stac_object()


def test_resolved_self_href() -> None:
    catalog = pystac.Catalog(id="test", description="test desc")
    with TemporaryDirectory() as temporary_directory:
        catalog.normalize_and_save(temporary_directory)
        path = os.path.join(temporary_directory, "catalog.json")
        catalog = pystac.Catalog.from_file(path)
        link = catalog.get_single_link(pystac.RelType.SELF)
        assert link
        link.resolve_stac_object()
        assert link.get_absolute_href() == make_posix_style(path)


def test_target_getter_setter(item: pystac.Item) -> None:
    link = pystac.Link("my rel", target="./foo/bar.json")
    assert link.target == "./foo/bar.json"
    assert link.get_target_str() == "./foo/bar.json"

    link.target = item
    assert link.target == item
    assert link.get_target_str() == item.get_self_href()

    link.target = "./bar/foo.json"
    assert link.target == "./bar/foo.json"


def test_get_target_str_no_href(item: pystac.Item) -> None:
    item.remove_links("self")
    link = pystac.Link("self", target=item)
    item.add_link(link)
    assert link.get_target_str() is None


def test_relative_self_href(item: pystac.Item) -> None:
    with TemporaryDirectory() as temporary_directory:
        pystac.write_file(
            item,
            include_self_link=False,
            dest_href=os.path.join(temporary_directory, "item.json"),
        )
        previous = os.getcwd()
        try:
            os.chdir(temporary_directory)
            item = cast(pystac.Item, pystac.read_file("item.json"))
            href = item.get_self_href()
            assert href
            assert os.path.isabs(href), f"Not an absolute path: {href}"
        finally:
            os.chdir(previous)


def test_auto_title_when_resolved(item: pystac.Item) -> None:
    extent = pystac.Extent.from_items([item])
    collection = pystac.Collection(
        id="my_collection",
        description="Test Collection",
        extent=extent,
        title="Collection Title",
    )
    link = pystac.Link("my rel", target=collection)

    assert collection.title == link.title


def test_auto_title_not_found(item: pystac.Item) -> None:
    extent = pystac.Extent.from_items([item])
    collection = pystac.Collection(
        id="my_collection",
        description="Test Collection",
        extent=extent,
    )
    link = pystac.Link("my rel", target=collection)

    assert link.title is None


def test_auto_title_is_serialized(item: pystac.Item) -> None:
    extent = pystac.Extent.from_items([item])
    collection = pystac.Collection(
        id="my_collection",
        description="Test Collection",
        extent=extent,
        title="Collection Title",
    )
    link = pystac.Link("my rel", target=collection)

    assert link.to_dict().get("title") == collection.title


def test_no_auto_title_if_not_resolved() -> None:
    link = pystac.Link("my rel", target="https://www.some-domain.com/path/to/thing.txt")

    assert link.title is None


def test_title_as_init_argument(item: pystac.Item) -> None:
    link_title = "Link title"
    extent = pystac.Extent.from_items([item])
    collection = pystac.Collection(
        id="my_collection",
        description="Test Collection",
        extent=extent,
        title="Collection Title",
    )
    link = pystac.Link("my rel", title=link_title, target=collection)

    assert link.title == link_title
    assert link.to_dict().get("title") == link_title


def test_serialize_link() -> None:
    href = "https://some-domain/path/to/item.json"
    title = "A Test Link"
    link = pystac.Link(pystac.RelType.SELF, href, pystac.MediaType.JSON, title)
    link_dict = link.to_dict()

    assert link_dict["rel"] == "self"
    assert link_dict["type"] == "application/json"
    assert link_dict["title"] == title
    assert link_dict["href"] == href


def test_static_from_dict_round_trip() -> None:
    test_cases: list[dict[str, Any]] = [
        {"rel": "", "href": ""},  # Not valid, but works.
        {"rel": "r", "href": "t"},
        {"rel": "r", "href": "/t"},
        {"rel": "r", "href": "t", "type": "a/b", "title": "t", "c": "d", "1": 2},
    ]
    for d in test_cases:
        d2 = pystac.Link.from_dict(d).to_dict()
        assert d == d2
    d = {"rel": "self", "href": "t"}
    d2 = {"rel": "self", "href": make_posix_style(os.path.join(os.getcwd(), "t"))}
    assert pystac.Link.from_dict(d).to_dict() == d2


def test_static_from_dict_failures() -> None:
    dicts: list[dict[str, Any]] = [{}, {"href": "t"}, {"rel": "r"}]
    for d in dicts:
        with pytest.raises(KeyError):
            pystac.Link.from_dict(d)


def test_static_collection(collection: pystac.Collection) -> None:
    link = pystac.Link.collection(collection)
    expected = {"rel": "collection", "href": None, "type": "application/json"}
    assert expected == link.to_dict()


def test_static_child(collection: pystac.Collection) -> None:
    link = pystac.Link.child(collection)
    expected = {"rel": "child", "href": None, "type": "application/json"}
    assert expected == link.to_dict()


def test_static_canonical_item(item: pystac.Item) -> None:
    link = pystac.Link.canonical(item)
    expected = {"rel": "canonical", "href": None, "type": "application/json"}
    assert expected == link.to_dict()


def test_static_canonical_collection(collection: pystac.Collection) -> None:
    link = pystac.Link.canonical(collection)
    expected = {"rel": "canonical", "href": None, "type": "application/json"}
    assert expected == link.to_dict()


class CustomLink(pystac.Link):
    pass


def test_inheritance_from_dict() -> None:
    link = CustomLink.from_dict(
        {"rel": "r", "href": "t", "type": "a/b", "title": "t", "c": "d", "1": 2}
    )
    assert isinstance(link, CustomLink)


def test_inheritance_collection(collection: Collection) -> None:
    link = CustomLink.collection(collection)
    assert isinstance(link, CustomLink)


def test_inheritance_child(collection: Collection) -> None:
    link = CustomLink.child(collection)
    assert isinstance(link, CustomLink)


def test_inheritance_canonical_item(item: Item) -> None:
    link = CustomLink.canonical(item)
    assert isinstance(link, CustomLink)


def test_inheritance_canonical_collection(collection: Collection) -> None:
    link = CustomLink.canonical(collection)
    assert isinstance(link, CustomLink)


def test_inheritance_clone() -> None:
    link = CustomLink.from_dict(
        {"rel": "r", "href": "t", "type": "a/b", "title": "t", "c": "d", "1": 2}
    )
    cloned_link = link.clone()
    assert isinstance(cloned_link, CustomLink)


def test_relative_self_link(tmp_path: Path) -> None:
    # https://github.com/stac-utils/pystac/issues/801
    item = Item("an-id", None, None, datetime.now(), {})
    item_as_dict = item.to_dict(include_self_link=False)
    item_as_dict["links"] = [{"href": "item.json", "rel": "self"}]
    item_as_dict["assets"] = {"data": {"href": "asset.tif"}}
    with open(tmp_path / "item.json", "w") as f:
        json.dump(item_as_dict, f)
    Path(tmp_path / "asset.tif").touch()
    collection = Collection("an-id", "a description", ARBITRARY_EXTENT)
    collection.add_link(Link("item", "item.json"))
    collection.save_object(
        include_self_link=False, dest_href=str(tmp_path / "collection.json")
    )
    collection = Collection.from_file(tmp_path / "collection.json")
    read_item = collection.get_item("an-id")
    assert read_item
    asset_href = read_item.assets["data"].get_absolute_href()
    assert asset_href
    assert Path(asset_href).exists()


@pytest.mark.parametrize("rel", HIERARCHICAL_LINKS)
def test_is_hierarchical(rel: str) -> None:
    assert Link(rel, "a-target").is_hierarchical()


@pytest.mark.parametrize(
    "rel", ["canonical", "derived_from", "alternate", "via", "prev", "next", "preview"]
)
def test_is_not_hierarchical(rel: str) -> None:
    assert not Link(rel, "a-target").is_hierarchical()


def test_item_link_type(item: Item) -> None:
    # https://github.com/stac-utils/pystac/issues/1494
    link = Link.item(item)
    assert link.media_type == "application/geo+json"



"""Tests for pystac.extensions.version."""

from collections.abc import Generator
from datetime import datetime

import pytest

from pystac import (
    Asset,
    Catalog,
    Collection,
    ExtensionNotImplemented,
    Extent,
    Item,
    SpatialExtent,
    TemporalExtent,
)
from pystac.errors import DeprecatedWarning, ExtensionTypeError
from pystac.extensions.version import (
    DEPRECATED,
    VERSION,
    VersionExtension,
    VersionRelType,
    ignore_deprecated,
)
from tests.utils import TestCases

URL_TEMPLATE: str = "http://example.com/catalog/%s.json"


@pytest.fixture
def item() -> Item:
    return make_item(2011)


@pytest.fixture
def item_from_file() -> Item:
    return Item.from_file(TestCases.get_path("data-files/version/item.json"))


@pytest.fixture
def collection() -> Collection:
    return make_collection(2011)


@pytest.fixture
def catalog() -> Catalog:
    return make_catalog(2011)


@pytest.fixture
def version() -> str:
    return "1.2.3"


def make_item(year: int) -> Item:
    """Create basic test items that are only slightly different."""
    asset_id = f"USGS/GAP/CONUS/{year}"
    start = datetime(year, 1, 2)

    item = Item(id=asset_id, geometry=None, bbox=None, datetime=start, properties={})
    item.set_self_href(URL_TEMPLATE % year)

    VersionExtension.add_to(item)

    return item


def make_collection(year: int) -> Collection:
    asset_id = f"my/collection/of/things/{year}"
    start = datetime(2014, 8, 10)
    end = datetime(year, 1, 3, 4, 5)
    bboxes = [[-180.0, -90.0, 180.0, 90.0]]
    spatial_extent = SpatialExtent(bboxes)
    intervals: list[list[datetime | None]] = [[start, end]]
    temporal_extent = TemporalExtent(intervals)
    extent = Extent(spatial_extent, temporal_extent)

    collection = Collection(asset_id, "desc", extent)
    collection.set_self_href(URL_TEMPLATE % year)

    VersionExtension.add_to(collection)

    return collection


def make_catalog(year: int) -> Catalog:
    """Create basic test catalogs that are only slightly different."""
    asset_id = f"USGS/GAP/CONUS/{year}"

    catalog = Catalog(id=asset_id, description=str(year))
    VersionExtension.add_to(catalog)

    return catalog


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match="^VersionExtension does not apply to type 'object'$"
    ):
        VersionExtension.ext(object())  # type: ignore


def test_rel_types() -> None:
    assert VersionRelType.LATEST.value == "latest-version"
    assert VersionRelType.PREDECESSOR.value == "predecessor-version"
    assert VersionRelType.SUCCESSOR.value == "successor-version"


def test_stac_extensions(item: Item) -> None:
    assert item.ext.has("version")


@pytest.mark.vcr()
def test_add_version(item: Item, version: str) -> None:
    item.ext.version.apply(version)
    assert version == item.ext.version.version
    assert DEPRECATED not in item.properties
    assert not item.ext.version.deprecated
    item.validate()


@pytest.mark.vcr()
def test_version_in_properties(item: Item, version: str) -> None:
    item.ext.version.apply(version, deprecated=True)
    assert VERSION in item.properties
    assert DEPRECATED in item.properties
    item.validate()


@pytest.mark.vcr()
def test_add_not_deprecated_version(item: Item, version: str) -> None:
    item.ext.version.apply(version, deprecated=False)
    assert DEPRECATED in item.properties
    assert not item.ext.version.deprecated
    item.validate()


@pytest.mark.vcr()
def test_add_deprecated_version(item: Item, version: str) -> None:
    item.ext.version.apply(version, deprecated=True)
    assert DEPRECATED in item.properties
    assert item.ext.version.deprecated
    item.validate()


@pytest.mark.vcr()
def test_latest(item: Item, version: str) -> None:
    year = 2013
    latest = make_item(year)
    item.ext.version.apply(version, latest=latest)
    assert item.ext.version.latest is latest
    expected_href = URL_TEMPLATE % year
    link = item.get_links(VersionRelType.LATEST)[0]
    assert expected_href == link.get_href()
    item.validate()


@pytest.mark.vcr()
def test_predecessor(item: Item, version: str) -> None:
    year = 2010
    predecessor = make_item(year)
    item.ext.version.apply(version, predecessor=predecessor)
    assert item.ext.version.predecessor is predecessor
    expected_href = URL_TEMPLATE % year
    link = item.get_links(VersionRelType.PREDECESSOR)[0]
    assert expected_href == link.get_href()
    item.validate()


@pytest.mark.vcr()
def test_successor(item: Item, version: str) -> None:
    year = 2012
    successor = make_item(year)
    item.ext.version.apply(version, successor=successor)
    assert item.ext.version.successor is successor
    expected_href = URL_TEMPLATE % year
    link = item.get_links(VersionRelType.SUCCESSOR)[0]
    assert expected_href == link.get_href()
    item.validate()


@pytest.mark.vcr()
def test_all_links(item: Item, version: str) -> None:
    deprecated = True
    latest = make_item(2013)
    predecessor = make_item(2010)
    successor = make_item(2012)
    item.ext.version.apply(
        version,
        deprecated=deprecated,
        latest=latest,
        predecessor=predecessor,
        successor=successor,
    )
    item.validate()


def test_full_copy(test_case_1_catalog: Catalog) -> None:
    # Fetch two items from the catalog
    item1 = next(test_case_1_catalog.get_items("area-1-1-imagery", recursive=True))
    item2 = next(test_case_1_catalog.get_items("area-2-2-imagery", recursive=True))

    # Enable the version extension on each, and link them
    # as if they are different versions of the same Item
    item1.ext.add("version")
    item2.ext.add("version")

    item1.ext.version.apply(version="2.0", predecessor=item2)
    item2.ext.version.apply(version="1.0", successor=item1, latest=item1)

    # Make a full copy of the catalog
    cat_copy = test_case_1_catalog.full_copy()

    # Retrieve the copied version of the items
    item1_copy = next(cat_copy.get_items("area-1-1-imagery", recursive=True))
    item2_copy = next(cat_copy.get_items("area-2-2-imagery", recursive=True))

    # Check to see if the version links point to the instances of the
    # item objects as they should.

    predecessor = item1_copy.get_single_link(VersionRelType.PREDECESSOR)
    assert predecessor is not None
    predecessor_target = predecessor.target
    successor = item2_copy.get_single_link(VersionRelType.SUCCESSOR)
    assert successor is not None
    successor_target = successor.target
    latest = item2_copy.get_single_link(VersionRelType.LATEST)
    assert latest is not None
    latest_target = latest.target

    assert predecessor_target is item2_copy
    assert successor_target is item1_copy
    assert latest_target is item1_copy


def test_setting_none_clears_link(item: Item, version: str) -> None:
    deprecated = False
    latest = make_item(2013)
    predecessor = make_item(2010)
    successor = make_item(2012)
    item.ext.version.apply(
        version, deprecated, latest=latest, predecessor=predecessor, successor=successor
    )

    item.ext.version.latest = None
    links = item.get_links(VersionRelType.LATEST)
    assert 0 == len(links)

    item.ext.version.predecessor = None
    links = item.get_links(VersionRelType.PREDECESSOR)
    assert 0 == len(links)

    item.ext.version.successor = None
    links = item.get_links(VersionRelType.SUCCESSOR)
    assert 0 == len(links)


def test_multiple_link_setting(item: Item, version: str) -> None:
    deprecated = False
    latest1 = make_item(2013)
    predecessor1 = make_item(2010)
    successor1 = make_item(2012)
    item.ext.version.apply(
        version,
        deprecated,
        latest=latest1,
        predecessor=predecessor1,
        successor=successor1,
    )

    year = 2015
    latest2 = make_item(year)
    expected_href = URL_TEMPLATE % year
    item.ext.version.latest = latest2
    links = item.get_links(VersionRelType.LATEST)
    assert 1 == len(links)
    assert expected_href == links[0].get_href()

    year = 2009
    predecessor2 = make_item(year)
    expected_href = URL_TEMPLATE % year
    item.ext.version.predecessor = predecessor2
    links = item.get_links(VersionRelType.PREDECESSOR)
    assert 1 == len(links)
    assert expected_href == links[0].get_href()

    year = 2014
    successor2 = make_item(year)
    expected_href = URL_TEMPLATE % year
    item.ext.version.successor = successor2
    links = item.get_links(VersionRelType.SUCCESSOR)
    assert 1 == len(links)
    assert expected_href == links[0].get_href()


def test_extension_not_implemented(item_from_file: Item) -> None:
    # Should raise exception if Item does not include extension URI
    item_from_file.stac_extensions.remove(VersionExtension.get_schema_uri())
    with pytest.raises(ExtensionNotImplemented):
        item_from_file.ext.version


def test_ext_add_to(item_from_file: Item) -> None:
    item_from_file.stac_extensions.remove(VersionExtension.get_schema_uri())
    item_from_file.ext.add("version")
    assert VersionExtension.get_schema_uri() in item_from_file.stac_extensions


def test_collection_stac_extensions(collection: Collection) -> None:
    assert collection.ext.has("version")


@pytest.mark.vcr()
def test_collection_add_version(collection: Collection, version: str) -> None:
    collection.ext.version.apply(version)
    assert collection.ext.version.version == version
    assert DEPRECATED not in collection.extra_fields
    collection.validate()


@pytest.mark.vcr()
def test_collection_validate_all(collection: Collection, version: str) -> None:
    deprecated = True
    latest = make_collection(2013)
    predecessor = make_collection(2010)
    successor = make_collection(2012)
    collection.ext.version.apply(
        version, deprecated, latest=latest, predecessor=predecessor, successor=successor
    )

    links = collection.get_links(VersionRelType.LATEST)
    assert 1 == len(links)
    links = collection.get_links(VersionRelType.PREDECESSOR)
    assert 1 == len(links)
    links = collection.get_links(VersionRelType.SUCCESSOR)
    assert 1 == len(links)

    collection.validate()


def test_catalog_stac_extensions(catalog: Catalog) -> None:
    assert catalog.ext.has("version")


@pytest.mark.vcr()
def test_catalog_add_version(catalog: Catalog, version: str) -> None:
    catalog.ext.version.apply(version)
    assert catalog.ext.version.version == version
    assert DEPRECATED not in catalog.extra_fields
    catalog.validate()


@pytest.mark.vcr()
def test_catalog_validate_all(catalog: Catalog, version: str) -> None:
    deprecated = True
    latest = make_collection(2013)
    predecessor = make_collection(2010)
    successor = make_collection(2012)
    catalog.ext.version.apply(
        version, deprecated, latest=latest, predecessor=predecessor, successor=successor
    )

    links = catalog.get_links(VersionRelType.LATEST)
    assert 1 == len(links)
    links = catalog.get_links(VersionRelType.PREDECESSOR)
    assert 1 == len(links)
    links = catalog.get_links(VersionRelType.SUCCESSOR)
    assert 1 == len(links)

    catalog.validate()


def test_item_deprecation_warning(
    item: Item, recwarn: Generator[pytest.WarningsRecorder, None, None]
) -> None:
    item.ext.version.deprecated = True
    item_dict = item.to_dict()
    with pytest.warns(DeprecatedWarning, match="The item '.*' is deprecated."):
        _ = Item.from_dict(item_dict)

    item.ext.version.deprecated = False
    item_dict = item.to_dict()
    _ = Item.from_dict(item_dict)
    assert len(list(recwarn)) == 0

    item.ext.version.deprecated = None
    item_dict = item.to_dict()
    _ = Item.from_dict(item_dict)
    assert len(list(recwarn)) == 0

    item.ext.remove("version")
    item_dict = item.to_dict()
    _ = Item.from_dict(item_dict)
    assert len(list(recwarn)) == 0


def test_collection_deprecation_warning(
    collection: Collection,
    recwarn: Generator[pytest.WarningsRecorder, None, None],
) -> None:
    collection.ext.version.deprecated = True
    collection_dict = collection.to_dict()
    with pytest.warns(DeprecatedWarning, match="The collection '.*' is deprecated."):
        _ = Collection.from_dict(collection_dict)

    collection.ext.version.deprecated = False
    collection_dict = collection.to_dict()
    _ = Collection.from_dict(collection_dict)
    assert len(list(recwarn)) == 0

    collection.ext.version.deprecated = None
    collection_dict = collection.to_dict()
    _ = Collection.from_dict(collection_dict)
    assert len(list(recwarn)) == 0

    collection.ext.remove("version")
    collection_dict = collection.to_dict()
    _ = Collection.from_dict(collection_dict)
    assert len(list(recwarn)) == 0


def test_ignore_deprecated_context_manager(
    item: Item, recwarn: Generator[pytest.WarningsRecorder, None, None]
) -> None:
    version = VersionExtension.ext(item, add_if_missing=True)
    version.deprecated = True
    item_dict = item.to_dict()
    with ignore_deprecated():
        _ = Item.from_dict(item_dict)
    assert len(list(recwarn)) == 0


def test_experimental(item: Item) -> None:
    # Added in v1.2.0
    assert item.ext.version.experimental is None
    assert "experimental" not in item.properties
    item.ext.version.experimental = True
    assert item.ext.version.experimental
    assert item.properties["experimental"]
    item.ext.version.experimental = False
    assert not item.properties["experimental"]
    item.ext.version.experimental = None
    assert "experimental" not in item.properties


@pytest.mark.vcr
def test_optional_version(item: Item) -> None:
    # Changed in v1.1.0
    assert item.ext.version.version is None
    assert "version" not in item.properties
    item.validate()
    item.ext.version.version = "final_final_2"
    assert "version" in item.properties
    item.ext.version.version = None
    assert "version" not in item.properties


@pytest.mark.vcr
def test_assets(item: Item) -> None:
    item.ext.remove("version")
    asset = Asset("example.tif")
    item.add_asset("data", asset)

    assert not asset.ext.has("version")
    asset.ext.add("version")
    assert asset.ext.has("version")
    assert item.ext.has("version")

    asset.ext.version.version = "final_final_2"
    assert asset.ext.version.version == "final_final_2"
    assert "version" in asset.extra_fields

    item.validate()


@pytest.mark.parametrize("version", ["v1.0.0", "v1.1.0"])
def test_migrate(version: str, item: Item) -> None:
    item_dict = item.to_dict(include_self_link=False, transform_hrefs=False)
    item_dict["stac_extensions"] = [
        f"https://stac-extensions.github.io/version/{version}/schema.json"
    ]
    item = Item.from_dict(item_dict, migrate=True)
    assert (
        "https://stac-extensions.github.io/version/v1.2.0/schema.json"
        in item.stac_extensions
    )



import pytest

import pystac
from pystac.cache import CollectionCache
from pystac.serialization import (
    identify_stac_object,
    identify_stac_object_type,
    merge_common_properties,
)
from pystac.serialization.identify import STACVersionID, STACVersionRange
from tests.utils import TestCases
from tests.utils.test_cases import ExampleInfo


class TestIdentify:
    @pytest.mark.parametrize("example", TestCases.get_examples_info())
    def test_identify(self, example: ExampleInfo) -> None:
        collection_cache = CollectionCache()
        path = example.path
        d = pystac.StacIO.default().read_json(path)
        if identify_stac_object_type(d) == pystac.STACObjectType.ITEM:
            merge_common_properties(
                d, json_href=path, collection_cache=collection_cache
            )

        actual = identify_stac_object(d)
        # Explicitly cover __repr__ functions in tests
        str_info = str(actual)
        assert isinstance(str_info, str)

        msg = f"Failed {path}:"

        assert actual.object_type == example.object_type, msg
        version_contained_in_range = actual.version_range.contains(example.stac_version)
        assert version_contained_in_range, msg
        assert set(actual.extensions) == set(example.extensions), msg

    def test_identify_non_stac_type(self) -> None:
        plain_feature_dict = {
            "type": "Feature",
            "properties": {},
            "geometry": {"type": "Point", "coordinates": [0, 0]},
        }

        assert identify_stac_object_type(plain_feature_dict) is None

    def test_identify_invalid_stac_object_with_version(self) -> None:
        # Has stac_version but is not a valid STAC object
        invalid_dict = {
            "id": "concepts",
            "title": "Concepts catalogs",
            "links": [
                {
                    "rel": "self",
                    "type": "application/json",
                    "href": "https://tamn.snapplanet.io/catalogs/concepts",
                },
                {
                    "rel": "root",
                    "type": "application/json",
                    "href": "https://tamn.snapplanet.io",
                },
            ],
            "stac_version": "1.1.0",
        }

        with pytest.raises(pystac.STACTypeError) as ctx:
            identify_stac_object(invalid_dict)

        assert "JSON (id = concepts) does not represent a STACObject instance." in str(
            ctx.value.args[0]
        )

    def test_identify_non_stac_raises_error(self) -> None:
        plain_feature_dict = {
            "type": "Feature",
            "properties": {},
            "geometry": {"type": "Point", "coordinates": [0, 0]},
        }

        with pytest.raises(pystac.STACTypeError) as ctx:
            identify_stac_object(plain_feature_dict)

        assert "JSON (id = unknown) does not represent a STACObject instance." in str(
            ctx.value.args[0]
        )

    def test_identify_invalid_with_stac_version(self) -> None:
        not_stac = {"stac_version": "0.9.0", "type": "Custom"}

        assert identify_stac_object_type(not_stac) is None


def test_version_ordering() -> None:
    assert STACVersionID("0.9.0") == STACVersionID("0.9.0")
    assert not STACVersionID("0.9.0") != STACVersionID("0.9.0")
    assert not STACVersionID("0.9.0") > STACVersionID("0.9.0")
    assert STACVersionID("1.0.0-beta.2") < "1.0.0"
    assert STACVersionID("0.9.1") > "0.9.0"
    assert not STACVersionID("0.9.0") > "0.9.0"
    assert STACVersionID("0.9.0") <= "0.9.0"
    assert STACVersionID("1.0.0-beta.1") <= STACVersionID("1.0.0-beta.2")
    assert not STACVersionID("1.0.0") < STACVersionID("1.0.0-beta.2")


def test_version_range_ordering() -> None:
    version_range = STACVersionRange("0.9.0", "1.0.0-beta.2")
    assert isinstance(str(version_range), str)
    assert version_range.contains("1.0.0-beta.1")
    assert not version_range.contains("1.0.0")
    assert version_range.is_later_than("0.8.9")

    version_range = STACVersionRange("0.9.0", "1.0.0-beta.1")
    assert not version_range.contains("1.0.0-beta.2")

    version_range = STACVersionRange(min_version="0.6.0-rc1", max_version="0.9.0")
    assert version_range.contains("0.9.0")


def test_version_range_set_to_single() -> None:
    version_range = STACVersionRange()
    version_range.set_min("1.0.0-beta.1")
    version_range.set_to_single("1.0.0")

    assert version_range.contains("1.0.0")


def test_version_range_set_min_and_max_directly() -> None:
    version_range = STACVersionRange()
    version_range.min_version = "1.0.0-beta.1"
    version_range.max_version = "1.1.0"

    assert version_range.contains("1.0.0")



from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    PathLike = os.PathLike[str]


def get_data_path(rel_path: str | PathLike) -> str:
    """Gets the absolute path to a file based on a path relative to the
    tests/data-files directory in this repo."""
    rel_path = os.fspath(rel_path)
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "tests", "data-files", rel_path)
    )



import json
from typing import Any, cast

from jsonschema import Draft7Validator, ValidationError

from pystac.errors import STACLocalValidationError
from pystac.version import STACVersion

VERSION = STACVersion.DEFAULT_STAC_VERSION


def _read_schema(file_name: str) -> dict[str, Any]:
    import importlib.resources

    with (
        importlib.resources.files("pystac.validation.jsonschemas")
        .joinpath(file_name)
        .open("r") as f
    ):
        return cast(dict[str, Any], json.load(f))


def get_local_schema_cache() -> dict[str, dict[str, Any]]:
    return {
        **{
            (
                f"https://schemas.stacspec.org/v{VERSION}/"
                f"{name}-spec/json-schema/{name}.json"
            ): _read_schema(f"stac-spec/v{VERSION}/{name}.json")
            for name in ("item", "catalog", "collection")
        },
        **{
            f"https://geojson.org/schema/{name}.json": _read_schema(
                f"geojson/{name}.json"
            )
            for name in ("Feature", "Geometry")
        },
        **{
            (
                f"https://schemas.stacspec.org/v{VERSION}/"
                f"item-spec/json-schema/{name}.json"
            ): _read_schema(f"stac-spec/v{VERSION}/{name}.json")
            for name in (
                "bands",
                "basics",
                "common",
                "data-values",
                "datetime",
                "instrument",
                "licensing",
                "provider",
            )
        },
    }


############################### DEPRECATED #################################

_deprecated_ITEM_SCHEMA_URI = (
    f"https://schemas.stacspec.org/v{VERSION}/item-spec/json-schema/item.json"
)
_deprecated_COLLECTION_SCHEMA_URI = (
    f"https://schemas.stacspec.org/v{VERSION}/"
    "collection-spec/json-schema/collection.json"
)
_deprecated_CATALOG_SCHEMA_URI = (
    f"https://schemas.stacspec.org/v{VERSION}/catalog-spec/json-schema/catalog.json"
)

deprecated_names = ["ITEM_SCHEMA_URI", "COLLECTION_SCHEMA_URI", "CATALOG_SCHEMA_URI"]


def __getattr__(name: str) -> Any:
    if name in deprecated_names:
        import warnings

        warnings.warn(f"{name} is deprecated and will be removed in v2.", FutureWarning)
        return globals()[f"_deprecated_{name}"]
    raise AttributeError(f"module {__name__} has no attribute {name}")


class LocalValidator:
    def __init__(self) -> None:
        """DEPRECATED"""
        import warnings

        warnings.warn(
            "``LocalValidator`` is deprecated and will be removed in v2.",
            DeprecationWarning,
        )
        self.schema_cache = get_local_schema_cache()

    def registry(self) -> Any:
        from referencing import Registry, Resource

        return Registry().with_resources(
            [(k, Resource.from_contents(v)) for k, v in self.schema_cache.items()]
        )

    def _validate_from_local(
        self, schema_uri: str, stac_dict: dict[str, Any]
    ) -> list[ValidationError]:
        if schema_uri == _deprecated_ITEM_SCHEMA_URI:
            validator = self.item_validator(VERSION)
        elif schema_uri == _deprecated_COLLECTION_SCHEMA_URI:
            validator = self.collection_validator(VERSION)
        elif schema_uri == _deprecated_CATALOG_SCHEMA_URI:
            validator = self.catalog_validator(VERSION)
        else:
            raise STACLocalValidationError(
                f"Schema not available locally: {schema_uri}"
            )
        return list(validator.iter_errors(stac_dict))

    def _validator(self, stac_type: str, version: str) -> Draft7Validator:
        schema = _read_schema(f"stac-spec/v{version}/{stac_type}.json")
        return Draft7Validator(schema, registry=self.registry())

    def catalog_validator(self, version: str = VERSION) -> Draft7Validator:
        return self._validator("catalog", version)

    def collection_validator(self, version: str = VERSION) -> Draft7Validator:
        return self._validator("collection", version)

    def item_validator(self, version: str = VERSION) -> Draft7Validator:
        return self._validator("item", version)



# TODO move all test case code to this file

import json
import shutil
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any

import pytest

from pystac import Asset, Catalog, Collection, Item, ItemCollection, Link

from .utils import ARBITRARY_BBOX, ARBITRARY_EXTENT, ARBITRARY_GEOM, TestCases

HERE = Path(__file__).resolve().parent


@pytest.fixture(scope="module")
def vcr_config() -> dict[str, Any]:
    def scrub_headers(response: dict[str, Any]) -> dict[str, Any]:
        retain = ["location"]
        response["headers"] = {
            key: value
            for (key, value) in response["headers"].items()
            if key.lower() in retain
        }
        return response

    return {"before_record_response": scrub_headers}


@pytest.fixture
def catalog() -> Catalog:
    return Catalog("test-catalog", "A test catalog")


@pytest.fixture
def collection() -> Catalog:
    return Collection("test-collection", "A test collection", ARBITRARY_EXTENT)


@pytest.fixture
def multi_extent_collection() -> Collection:
    # TODO this code is repeated many times; refactor to use this fixture
    return Collection.from_file(
        TestCases.get_path("data-files/collections/multi-extent.json")
    )


@pytest.fixture
def item() -> Item:
    return Item("test-item", ARBITRARY_GEOM, ARBITRARY_BBOX, datetime.now(), {})


@pytest.fixture
def asset(item: Item) -> Asset:
    item.add_asset("foo", Asset("https://example.tif"))
    return item.assets["foo"]


@pytest.fixture
def link(item: Item) -> Link:
    item.add_link(Link(rel="child", target="https://example.tif"))
    return item.links[0]


@pytest.fixture
def test_case_1_catalog() -> Catalog:
    return TestCases.case_1()


@pytest.fixture
def test_case_8_collection() -> Collection:
    return TestCases.case_8()


@pytest.fixture
def projection_landsat8_item() -> Item:
    path = TestCases.get_path("data-files/projection/example-landsat8.json")
    return Item.from_file(path)


def get_data_file(rel_path: str) -> str:
    return str(HERE / "data-files" / rel_path)


@pytest.fixture
def sample_item_dict() -> dict[str, Any]:
    m = TestCases.get_path("data-files/item/sample-item.json")
    with open(m) as f:
        item_dict: dict[str, Any] = json.load(f)
    return item_dict


@pytest.fixture
def sample_item() -> Item:
    return Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))


@pytest.fixture
def sample_item_collection() -> ItemCollection:
    return ItemCollection.from_file(
        TestCases.get_path("data-files/item-collection/sample-item-collection.json")
    )


@pytest.fixture
def sample_items(sample_item_collection: ItemCollection) -> list[Item]:
    return list(sample_item_collection)


@pytest.fixture(scope="function")
def tmp_asset(tmp_path: Path) -> Asset:
    """Copy the entirety of test-case-2 to tmp and"""
    src = get_data_file("catalogs/test-case-2")
    dst = str(tmp_path / str(uuid.uuid4()))
    shutil.copytree(src, dst)

    catalog = Catalog.from_file(f"{dst}/catalog.json")
    item = next(catalog.get_items(recursive=True))
    return next(v for v in item.assets.values())


@pytest.fixture(autouse=True)
def clear_validator() -> None:
    from pystac.validation import RegisteredValidator

    RegisteredValidator._validator = None
